#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : plotly_grapher.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019


import  textwrap

import  numpy                   as          np

import  plotly.graph_objects    as          go
from    plotly.subplots         import      make_subplots

import  pept


class PlotlyGrapher:
    '''A class for PEPT data visualisation using Plotly-based 3D graphs.

    The **PlotlyGrapher** class can create and automatically configure an
    arbitrary number of 3D subplots for PEPT data visualisation. They are by
    default set to use the *alternative PEPT 3D axes convention* - having the
    *y*-axis pointing upwards, such that the vertical screens of a PEPT scanner
    represent the *xy*-plane.

    This class can be used to draw 3D scatter or line plots, with optional
    colour-coding using extra data columns (e.g. relative tracer activity or
    trajectory label).

    It also provides easy access to the most common configuration parameters
    for the plots, such as axes limits, subplot titles, colorbar titles, etc.
    It can work with pre-computed Plotly traces (such as the ones from the
    `pept` base classes), as well as with numpy arrays.

    Attributes
    ----------
    xlim : list or numpy.ndarray
        A list of length 2, formatted as `[x_min, x_max]`, where `x_min` is
        the lower limit of the x-axis of all the subplots and `x_max` is the
        upper limit of the x-axis of all the subplots.

    ylim : list or numpy.ndarray
        A list of length 2, formatted as `[y_min, y_max]`, where `y_min` is
        the lower limit of the y-axis of all the subplots and `y_max` is the
        upper limit of the y-axis of all the subplots.

    zlim : list or numpy.ndarray
        A list of length 2, formatted as `[z_min, z_max]`, where `z_min` is
        the lower limit of the z-axis of all the subplots and `z_max` is the
        upper limit of the z-axis of all the subplots.

    fig : Plotly.Figure instance
        A Plotly.Figure instance, with any number of subplots (as defined by
        `rows` and `cols`) pre-configured for PEPT data.

    Methods
    -------
    create_figure()
        Create a Plotly figure, pre-configured for PEPT data.

    add_points(points, row = 1, col = 1, size = 2.0, color = None,\
               opacity = 0.8, colorbar = True, colorbar_col = -1,\
               colorscale = "Magma", colorbar_title = None)
        Create and plot a trace for all the points in a numpy array or
        `pept.PointData`, with possible color-coding.

    add_lines(lines, row = 1, col = 1, width = 2.0, color = None,\
              opacity = 0.6, colorbar = True, colorbar_col = 0,\
              colorscale = "Magma", colorbar_title = None)
        Create and plot a trace for all the lines in a numpy array or
        `pept.LineData`, with possible color-coding.

    add_pixels(pixels, row = 1, col = 1, condition = lambda pixels: pixels > 0,
               opacity = 0.9, colorscale = "Magma")
        Create and plot a trace with all the pixels in this class, with
        possible filtering.

    add_voxels(voxels, row = 1, col = 1, condition = lambda voxels: voxels > 0,
               size = 4, color = None, opacity = 0.4, colorbar = True,
               colorscale = "Magma", colorbar_title = None)
        Create and plot a trace for all the voxels in a `pept.Voxels` or
        `pept.VoxelData` instance, with possible filtering.

    add_trace(trace, row = 1, col = 1)
        Add a precomputed Plotly trace to a given subplot.

    add_traces(traces, row = 1, col = 1)
        Add a list of precomputed Plotly traces to a given subplot.

    equalise_axes()
        Equalise the axes' limits of all subplots.

    to_html(filepath, equal_axes = True, include_plotlyjs = True)
        Save the current Plotly figure as a self-contained HTML webpage.

    show(equal_axes = True)
        Show the Plotly figure, optionally setting equal axes limits.

    Raises
    ------
    ValueError
        If `xlim`, `ylim` or `zlim` are not lists of length 2.

    Examples
    --------
    The figure is created when instantiating the class.

    >>> grapher = PlotlyGrapher()
    >>> lors = LineData(raw_lors...)        # Some example lines
    >>> points = PointData(raw_points...)   # Some example points

    Using pre-computed traces from the `LineData` and `PointData` classes:

    >>> grapher.add_trace(lors.lines_trace())
    >>> grapher.add_traces([lors.lines_trace(), points.points_trace()])

    Creating a trace based on a numpy array:

    >>> sample_lors = lors[0]           # A numpy array of a single sample
    >>> sample_points = points[0]
    >>> grapher.add_lines(sample_lors)
    >>> grapher.add_points(sample_points)

    Showing the plot:

    >>> grapher.show()

    If you'd like to show the plot in your browser, you can set the default
    Plotly renderer:

    >>> plotly.io.renderers.default = "browser"

    More examples are given in the docstrings of the `add_points`, `add_lines`
    methods.
    '''

    def __init__(
        self,
        rows = 1,
        cols = 1,
        xlim = None,
        ylim = None,
        zlim = None,
        subplot_titles = ["  "]
    ):
        '''`PlotlyGrapher` class constructor.

        Parameters
        ----------
        rows : int, optional
            The number of rows of subplots. The default is 1.

        cols : int, optional
            The number of columns of subplots. The default is 1.

        xlim : list or numpy.ndarray, optional
            A list of length 2, formatted as `[x_min, x_max]`, where `x_min` is
            the lower limit of the x-axis of all the subplots and `x_max` is
            the upper limit of the x-axis of all the subplots.

        ylim : list or numpy.ndarray, optional
            A list of length 2, formatted as `[y_min, y_max]`, where `y_min` is
            the lower limit of the y-axis of all the subplots and `y_max` is
            the upper limit of the y-axis of all the subplots.

        zlim : list or numpy.ndarray, optional
            A list of length 2, formatted as `[z_min, z_max]`, where `z_min` is
            the lower limit of the z-axis of all the subplots and `z_max` is
            the upper limit of the z-axis of all the subplots.

        subplot_titles : list of str, default ["  "]
            A list of the titles of the subplots - e.g. ["plot a)", "plot b)"].
            The default is a list of empty strings.

        Raises
        ------
        ValueError
            If `rows` < 1 or `cols` < 1.

        ValueError
            If `xlim`, `ylim` or `zlim` are not lists of length 2.
        '''

        rows = int(rows)
        cols = int(cols)

        if rows < 1 or cols < 1:
            raise ValueError((
                "\n[ERROR]: The number of rows and cols have to be larger "
                f"than 1. Received rows = {rows}; cols = {cols}.\n"
            ))

        self._rows = rows
        self._cols = cols

        if xlim is not None:
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or xlim.shape[0] != 2:
                raise ValueError((
                    "\n[ERROR]: xlim needs to be a list of length 2, formatted"
                    f" as xlim = [x_min, x_max]. Received {xlim}.\n"
                ))

        if ylim is not None:
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or ylim.shape[0] != 2:
                raise ValueError((
                    "\n[ERROR]: ylim needs to be a list of length 2, formatted"
                    f" as ylim = [y_min, y_max]. Received {ylim}.\n"
                ))

        if zlim is not None:
            zlim = np.asarray(zlim, dtype = float)

            if zlim.ndim != 1 or zlim.shape[0] != 2:
                raise ValueError((
                    "\n[ERROR]: zlim needs to be a list of length 2, formatted"
                    f" as zlim = [z_min, z_max]. Received {zlim}.\n"
                ))

        self._xlim = xlim
        self._ylim = ylim
        self._zlim = zlim

        self._subplot_titles = subplot_titles
        # Pad the subplot titles that were not set with empty strings.
        self._subplot_titles.extend(['  '] * (rows * cols -
                                              len(subplot_titles)))

        self._fig = self.create_figure()


    def create_figure(self):
        '''Create a Plotly figure, pre-configured for PEPT data.

        This function creates a Plotly figure with an arbitrary number of
        subplots, as given in the class instantiation call. It configures them
        to have the *y*-axis pointing upwards, as per the PEPT 3D axes
        convention. It also sets the axes limits and labels.

        Returns
        -------
        fig : Plotly Figure instance
            A Plotly Figure instance, with any number of subplots (as defined
            when instantiating the class) pre-configured for PEPT data.
        '''

        specs = [[{"type": "scatter3d"}] * self._cols] * self._rows

        self._fig = make_subplots(
            rows = self._rows,
            cols = self._cols,
            specs = specs,
            subplot_titles = self._subplot_titles,
            horizontal_spacing = 0.005,
            vertical_spacing = 0.05
        )

        self._fig['layout'].update(
            margin = dict(l = 0, r = 0, b = 30, t = 30),
            showlegend = False
        )

        # For every subplot (scene), set axes' ratios and limits
        # Also set the y axis to point upwards
        # Plotly naming convention of scenes: 'scene', 'scene2', etc.
        for i in range(self._rows):
            for j in range(self._cols):
                if i == j == 0:
                    scene = "scene"
                else:
                    scene = "scene{}".format(i * self._cols + j + 1)

                # Justify subplot title on the left
                # self._fig.layout.annotations[i * self._cols + j].update(
                #     x = (j + 0.15) / self._cols
                # )

                self._fig["layout"][scene].update(
                    xaxis = dict(
                        range = self._xlim,
                        title = dict(text = "<i>x</i> (mm)")
                    ),
                    yaxis = dict(
                        range = self._ylim,
                        title = dict(text = "<i>y</i> (mm)")
                    ),
                    zaxis = dict(
                        range = self._zlim,
                        title = dict(text = "<i>z</i> (mm)")
                    ),
                    aspectmode = "manual",
                    aspectratio = dict(x = 1, y = 1, z = 1),
                    camera = dict(
                        up = dict(x = 0, y = 1, z = 0),
                        eye = dict(x = 1, y = 1, z = 1)
                    ),
                )

        return self._fig


    @property
    def xlim(self):
        return self._xlim


    @xlim.setter
    def xlim(self, xlim):
        xlim = np.asarray(xlim, dtype = float)

        if xlim.ndim != 1 or xlim.shape[0] != 2:
            raise ValueError((
                "\n[ERROR]: xlim needs to be a list of length 2, formatted as"
                f"xlim = [x_min, x_max]. Received {xlim}.\n"
            ))

        self._xlim = xlim

        # For every subplot (scene), update axes' limits
        # Plotly naming convention of scenes: 'scene', 'scene2', etc.
        for i in range(self._rows):
            for j in range(self._cols):
                if i == j == 0:
                    scene = 'scene'
                else:
                    scene = 'scene{}'.format(i * self._cols + j + 1)

                self._fig['layout'][scene].update(
                    xaxis = dict(range = self._xlim)
                )


    @property
    def ylim(self):
        return self._ylim


    @ylim.setter
    def ylim(self, ylim):
        ylim = np.asarray(ylim, dtype = float)

        if ylim.ndim != 1 or ylim.shape[0] != 2:
            raise ValueError((
                "\n[ERROR]: ylim needs to be a list of length 2, formatted as"
                f"ylim = [y_min, y_max]. Received {ylim}.\n"
            ))

        self._ylim = ylim

        # For every subplot (scene), update axes' limits
        # Plotly naming convention of scenes: 'scene', 'scene2', etc.
        for i in range(self._rows):
            for j in range(self._cols):
                if i == j == 0:
                    scene = 'scene'
                else:
                    scene = 'scene{}'.format(i * self._cols + j + 1)

                self._fig['layout'][scene].update(
                    yaxis = dict(range = self._ylim)
                )


    @property
    def zlim(self):
        return self._zlim


    @zlim.setter
    def zlim(self, zlim):
        zlim = np.asarray(zlim, dtype = float)

        if zlim.ndim != 1 or zlim.shape[0] != 2:
            raise ValueError((
                "\n[ERROR]: zlim needs to be a list of length 2, formatted as"
                f"zlim = [z_min, z_max]. Received {zlim}.\n"
            ))

        self._zlim = zlim

        # For every subplot (scene), update axes' limits
        # Plotly naming convention of scenes: 'scene', 'scene2', etc.
        for i in range(self._rows):
            for j in range(self._cols):
                if i == j == 0:
                    scene = 'scene'
                else:
                    scene = 'scene{}'.format(i * self._cols + j + 1)

                self._fig['layout'][scene].update(
                    zaxis = dict(range = self._zlim)
                )


    @property
    def fig(self):
        return self._fig


    def add_points(
        self,
        points,
        row = 1,
        col = 1,
        size = 2.0,
        color = None,
        opacity = 0.8,
        colorbar = True,
        colorbar_col = -1,
        colorscale = "Magma",
        colorbar_title = None
    ):
        '''Create and plot a trace for all the points in a numpy array or
        `pept.PointData`, with possible color-coding.

        Creates a `plotly.graph_objects.Scatter3d` object for all the points
        included in the numpy array or `pept.PointData` instance (or subclass
        thereof!) `points` and adds it to the subplot determined by `row` and
        `col`.

        The expected data row is [time, x1, y1, z1, ...].

        Parameters
        ----------
        points : (M, N >= 4) numpy.ndarray or pept.PointData
            The expected data columns are: [time, x1, y1, z1, etc.]. If a
            `pept.PointData` instance (or subclass thereof) is received, the
            inner `points` will be used.

        row : int, default 1
            The row of the subplot to add a trace to.

        col : int, default 1
            The column of the subplot to add a trace to.

        size : float, default 2.0
            The marker size of the points.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.8
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the data in the `points` column
            `colorbar_col`. Is overridden by `color` if set.

        colorbar_col : int, default -1
            The column in `points` that will be used to color the points. Only
            has an effect if `colorbar` is set to True. The default is -1 (the
            last column).

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the `colorbar_col` column in the
            input data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above it.

        Raises
        ------
        ValueError
            If `points` is not a numpy.ndarray with shape (M, N), where N >= 4.

        Notes
        -----
        If a colorbar is to be used (i.e. `colorbar = True` and `color = None`)
        and there are fewer than 10 unique values in the `colorbar_col` column
        in `points`, then the points for each unique label will be added as
        separate traces.

        This is helpful for cases such as when plotting points with labelled
        trajectories, as when there are fewer than 10 trajectories, the
        distinct colours automatically used by Plotly when adding multiple
        traces allow the points to be better distinguished.

        Examples
        --------
        Add an array of points (data columns: [time, x, y, z]) to a
        `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> points_raw = np.array(...)      # shape (N, M >= 4)
        >>> grapher.add_points(points_raw)
        >>> grapher.show()

        Add all the points in a `PointData` instance:

        >>> point_data = pept.PointData(...)    # Some example data
        >>> grapher.add_points(point_data)
        >>> grapher.show()

        Note that the above method can only add the whole `points` attribute of
        `PointData`. If you'd like to only plot some samples, use the
        `PointData.points_trace([sample_indices])` method:

        >>> trace = point_data.points_trace([0, 1, 2]) # Select samples 0, 1, 2
        >>> grapher.add_trace(trace)

        If you have an extremely large number of points in a numpy array, you
        can plot every 10th point using slices:

        >>> pts = np.array(...)         # shape (N, M >= 4), N very large
        >>> grapher.add_points(pts[::10])

        '''

        # If a pept.PointData instance (or subclass thereof!) is received, just
        # take the inner `points`. Otherwise treat it as an array.
        if isinstance(points, pept.PointData):
            points = points.points
        else:
            points = np.asarray(points, dtype = float)

        # Check that points has shape (M, 4)
        if points.ndim != 2 or points.shape[1] < 4:
            raise ValueError((
                "\n[ERROR]: `points` should have dimensions (M, N), where "
                "N >= 4. Received {}\n".format(points.shape)
            ))

        # No need to type-check the other parameters as Plotly will do that
        # anyway...

        # Create the dictionary of marker properties
        marker = dict(
            size = size,
            color = color,
            opacity = opacity
        )

        # Update `marker` if a colorbar is requested AND color is None.
        if colorbar and color is None:
            marker.update(colorscale = colorscale)
            if colorbar_title is not None:
                marker["colorbar"] = dict(title = colorbar_title)

            # Special case: if there are less than 10 values in the colorbar
            # column, add them as separate traces for better distinction
            # between colours.
            labels = np.unique(points[:, colorbar_col])

            if len(labels) <= 10:
                for label in labels:
                    selected = points[points[:, colorbar_col] == label]

                    self._fig.add_trace(
                        go.Scatter3d(
                            x = selected[:, 1],
                            y = selected[:, 2],
                            z = selected[:, 3],
                            mode = "markers",
                            marker = marker
                        ),
                        row = row,
                        col = col
                    )
                return

            # Otherwise just use a typical continuous colorbar for all the
            # values in colorbar_col.
            else:
                marker['color'] = points[:, colorbar_col]

        coords_x = points[:, 1]
        coords_y = points[:, 2]
        coords_z = points[:, 3]

        trace = go.Scatter3d(
            x = coords_x,
            y = coords_y,
            z = coords_z,
            mode = "markers",
            marker = marker
        )

        self._fig.add_trace(trace, row = row, col = col)


    def add_lines(
        self,
        lines,
        row = 1,
        col = 1,
        width = 2.0,
        color = None,
        opacity = 0.6,
        colorbar = True,
        colorbar_col = 0,
        colorscale = "Magma",
        colorbar_title = None
    ):
        '''Create and plot a trace for all the lines in a numpy array or
        `pept.LineData`, with possible color-coding.

        Creates a `plotly.graph_objects.Scatter3d` object for all the lines
        included in the numpy array or `pept.LineData` instance (or subclass
        thereof!) `lines` and adds it to the subplot determined by `row` and
        `col`.

        It expects LoR-like data, where each line is defined by two points. The
        expected data columns are [time, x1, y1, z1, x2, y2, z2, ...].

        Parameters
        ----------
        lines : (M, N >= 7) numpy.ndarray or pept.LineData
            The expected data columns: [time, x1, y1, z1, x2, y2, z2, etc.]. If
            a `pept.LineData` instance (or subclass thereof) is received, the
            inner `lines` will be used.

        row : int, default 1
            The row of the subplot to add a trace to.

        col : int, default 1
            The column of the subplot to add a trace to.

        width : float, default 2.0
            The width of the lines.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.6
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the data in the `lines` column
            `colorbar_col`. Is overridden if `color` is set. The default is
            True, so that every line has a different color.

        colorbar_col : int, default 0
            The column in the data samples that will be used to color the
            points. Only has an effect if `colorbar` is set to True. The
            default is 0 (the first column - time).

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the `colorbar_col` column in the
            input data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above it.

        Raises
        ------
        ValueError
            If `lines` is not a numpy.ndarray with shape (M, N), where N >= 7.

        Examples
        --------
        Add an array of lines (data columns: [t, x1, y1, z1, x2, y2, z2]) to a
        `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines_raw = np.array(...)           # shape (N, M >= 7)
        >>> grapher.add_lines(lines_raw)
        >>> grapher.show()

        Add all the lines in a `LineData` instance:

        >>> line_data = pept.LineData(...)      # Some example data
        >>> grapher.add_lines(line_data)
        >>> grapher.show()

        Note that the above method can only add the whole `lines` attribute of
        `LineData`. If you'd like to only plot some samples, use the
        `LineData.lines_trace([sample_indices])` method:

        >>> trace = line_data.lines_trace([0, 1, 2]) # Select samples 0, 1, 2
        >>> grapher.add_trace(trace)

        If you have a very large number of lines in a numpy array, you can plot
        every 10th point using slices:

        >>> lines_raw = np.array(...)       # shape (N, M >= 7), N very large
        >>> grapher.add_lines(lines_raw[::10])

        '''

        # If a pept.LineData instance (or subclass thereof!) is received, just
        # take the inner `lines`. Otherwise treat it as an array.
        if isinstance(lines, pept.LineData):
            lines = lines.lines
        else:
            lines = np.asarray(lines, dtype = float)

        # Check that lines has shape (N, 7)
        if lines.ndim != 2 or lines.shape[1] < 7:
            raise ValueError((
                "\n[ERROR]: `lines` should have dimensions (M, N), where "
                "N >= 7. Received {}\n".format(lines.shape)
            ))

        marker = dict(
            width = width,
            color = color,
        )

        if colorbar:
            if color is None:
                marker['color'] = []

            marker.update(colorscale = colorscale)
            if colorbar_title is not None:
                marker.update(colorbar = dict(title = colorbar_title))

        coords_x = []
        coords_y = []
        coords_z = []

        for line in lines:
            coords_x.extend([line[1], line[4], None])
            coords_y.extend([line[2], line[5], None])
            coords_z.extend([line[3], line[6], None])

            if colorbar and color is None:
                marker['color'].extend(3 * [line[colorbar_col]])

        coords_x = np.array(coords_x, dtype = float)
        coords_y = np.array(coords_y, dtype = float)
        coords_z = np.array(coords_z, dtype = float)

        trace = go.Scatter3d(
            x = coords_x,
            y = coords_y,
            z = coords_z,
            mode = 'lines',
            opacity = opacity,
            line = marker
        )

        self._fig.add_trace(trace, row = row, col = col)


    def add_pixels(
        self,
        pixels,
        row = 1,
        col = 1,
        condition = lambda pixels: pixels > 0,
        opacity = 0.9,
        colorscale = "Magma",
    ):
        '''Create and plot a trace with all the pixels in this class, with
        possible filtering.

        Creates a `plotly.graph_objects.Surface` object for the centres of
        all pixels encapsulated in a `pept.Pixels` instance, colour-coding the
        pixel value.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all pixels that have a value larger
        than 0.

        Parameters
        ----------
        voxels : pept.Pixels
            The pixel space, encapsulated in a `pept.Pixels` instance (or
            subclass thereof). Only `pept.Pixels` are accepted as raw pixels on
            their own do not contain data about the spatial coordinates of the
            pixel box.

        row : int, default 1
            The row of the subplot to add a trace to.

        col : int, default 1
            The column of the subplot to add a trace to.

        condition : function, default `lambda pixels: pixels > 0`
            The filtering function applied to the pixel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all pixels that
            should be plotted. The default, `lambda x: x > 0` selects all
            pixels which have a value larger than 0.

        opacity : float, default 0.4
            The opacity of the surface, where 0 is transparent and 1 is fully
            opaque.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        Examples
        --------
        Pixellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines = np.array(...)                   # shape (N, M >= 7)
        >>> lines2d = lines[:, [0, 1, 2, 4, 5]]     # select x, y of lines
        >>> number_of_pixels = [10, 10]
        >>> pixels = pept.Pixels.from_lines(lines2d, number_of_pixels)
        >>> grapher.add_lines(lines)
        >>> grapher.add_trace(pixels.pixels_trace())
        >>> grapher.show()

        '''

        if not isinstance(pixels, pept.Pixels):
            raise TypeError(textwrap.fill((
                "The input `pixels` must be an instance of `pept.Pixels` (or "
                f"subclass thereof. Received {type(pixels)}."
            )))

        trace = pixels.pixels_trace(
            condition = condition,
            opacity = opacity,
            colorscale = colorscale,
        )

        self._fig.add_trace(trace, row = row, col = col)


    def add_voxels(
        self,
        voxels,
        row = 1,
        col = 1,
        condition = lambda voxel_data: voxel_data > 0,
        size = 4,
        color = None,
        opacity = 0.4,
        colorbar = True,
        colorscale = "Magma",
        colorbar_title = None,
    ):
        '''Create and plot a trace for all the voxels in a `pept.Voxels` or
        `pept.VoxelData` instance, with possible filtering.

        Creates a `plotly.graph_objects.Scatter3d` object for the centres of
        all voxels encapsulated in a `pept.Voxels` instance, colour-coding the
        voxel value. The trace is added to the subplot determined by `row` and
        `col`.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all voxels that have a value larger
        than 0.

        Parameters
        ----------
        voxels : pept.Voxels or pept.VoxelData
            The voxel space, encapsulated in a `pept.Voxels` or
            `pept.VoxelData` instance (or subclass thereof). If a `VoxelData`
            is received, all the voxels will be accumulated / superimpoesd.
            Only these classes are accepted as raw voxels on their own do not
            contain data about the spatial coordinates of the voxel box.

        row : int, default 1
            The row of the subplot to add a trace to.

        col : int, default 1
            The column of the subplot to add a trace to.

        condition : function, default `lambda voxel_data: voxel_data > 0`
            The filtering function applied to the voxel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all voxels that
            should be plotted. The default, `lambda x: x > 0` selects all
            voxels which have a value larger than 0.

        size : float, default 4
            The size of the plotted voxel points. Note that due to the large
            number of voxels in typical applications, the *voxel centres* are
            plotted as square points, which provides an easy to understand
            image that is also fast and responsive.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.4
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the voxel values. Is overridden if
            `color` is set.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above it.

        Raises
        ------
        TypeError
            If `voxels` is not an instance of `pept.Voxels` or subclass
            thereof.

        Examples
        --------
        Voxellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines = np.array(...)           # shape (N, M >= 7)
        >>> number_of_voxels = [10, 10, 10]
        >>> voxels = pept.Voxels(lines, number_of_voxels)
        >>> grapher.add_lines(lines)
        >>> grapher.add_voxels(voxels)
        >>> grapher.show()

        '''

        if isinstance(voxels, pept.VoxelData):
            voxels = voxels.accumulate(verbose = False)
        elif not isinstance(voxels, pept.Voxels):
            raise TypeError(textwrap.fill((
                "The input `voxels` must be an instance of `pept.Voxels` or "
                "`pept.VoxelData` (or subclass thereof). Received "
                f"{type(voxels)}."
            )))

        trace = voxels.voxels_trace(
            condition = condition,
            size = size,
            color = color,
            opacity = opacity,
            colorbar = colorbar,
            colorscale = colorscale,
            colorbar_title = colorbar_title,
        )

        self._fig.add_trace(trace, row = row, col = col)


    def add_trace(self, trace, row = 1, col = 1):
        '''Add a precomputed Plotly trace to a given subplot.

        The equivalent of the Plotly figure.add_trace method.

        Parameters
        ----------
        trace : Plotly trace (Scatter3d)
            A precomputed Plotly trace

        row : int, default 1
            The row of the subplot to add a trace to.

        col : int, default 1
            The column of the subplot to add a trace to.
        '''

        # Add precomputed trace
        self._fig.add_trace(trace, row = row, col = col)


    def add_traces(self, traces, row = 1, col = 1):
        '''Add a list of precomputed Plotly traces to a given subplot.

        The equivalent of the Plotly figure.add_traces method.

        Parameters
        ----------
        traces : list [ Plotly trace (Scatter3d) ]
            A list of precomputed Plotly traces

        row : int, default 1
            The row of the subplot to add the traces to.

        col : int, default 1
            The column of the subplot to add the traces to.
        '''

        # Add precomputed traces
        self._fig.add_traces(
            traces,
            rows = [row] * len(traces),
            cols = [col] * len(traces)
        )


    def equalise_axes(self):
        '''Equalise the axes of all subplots by setting the system limits
        `xlim` and `ylim` to equal values, such that all data plotted is
        within the plotted bounds.
        '''
        # Compute min, max for the `x`, `y`, `z` dimensions for every
        # dataset added to `_fig`
        def get_min_max(fig_data):
            # Convert x, y, z attributes of `fig_data` to numpy arrays with
            # `dtype = float`, such that `None` entries are casted to
            # np.nan. Then find min, max for each dimension.
            x = np.asarray(fig_data.x, dtype = float)
            y = np.asarray(fig_data.y, dtype = float)
            z = np.asarray(fig_data.z, dtype = float)

            # Find min, max, ignoring np.nans
            xmin = np.nanmin(x)
            xmax = np.nanmax(x)

            ymin = np.nanmin(y)
            ymax = np.nanmax(y)

            zmin = np.nanmin(z)
            zmax = np.nanmax(z)

            return [xmin, xmax, ymin, ymax, zmin, zmax]

        # `lims` columns: [xmin, xmax, ymin, ymax, zmin, zmax].
        lims = [get_min_max(fig_data) for fig_data in self._fig.data]
        lims = np.array(lims, order = "F")

        # Find global min and max for each dimension.
        mins = lims[:, [0, 2, 4]].min(axis = 0)
        maxs = lims[:, [1, 3, 5]].max(axis = 0)

        # Find greatest range in all dimensions.
        max_range = (maxs - mins).max()

        # Find mean for each dimension to centre plot around it.
        mean = (maxs + mins) / 2

        # Finally, set xlim, ylim, zlim to be centred around their mean,
        # with a span of max_range.
        self.xlim = [mean[0] - max_range / 2, mean[0] + max_range / 2]
        self.ylim = [mean[1] - max_range / 2, mean[1] + max_range / 2]
        self.zlim = [mean[2] - max_range / 2, mean[2] + max_range / 2]


    def show(self, equal_axes = True):
        '''Show the Plotly figure, optionally setting equal axes limits.

        Note that the figure will be shown on the Plotly-configured renderer
        (e.g. browser, or PDF). The available renderers can be found by running
        the following code:

        >>> import plotly.io as pio
        >>> pio.renderers

        If you want an interactive figure in the browser, run the following:

        >>> pio.renderers.default = "browser"

        Parameters
        ----------
        equal_axes : bool, default True
            Set `xlim`, `ylim`, `zlim` to equal ranges such that the axes
            limits are equalised. Only has an effect if `xlim`, `ylim` and
            `zlim` are all `None`. If `False`, the default Plotly behaviour is
            used (i.e. automatically use min, max for each dimension).
        '''

        if (equal_axes is True and self.xlim is None and self.ylim is None and
                self.zlim is None):
            self.equalise_axes()

        self._fig.show()


    def to_html(
        self,
        filepath,
        equal_axes = True,
        include_plotlyjs = True,
    ):
        '''Save the current Plotly figure as a self-contained HTML webpage.

        Parameters
        ----------
        filepath : str or writeable
            Path or open file descriptor to save the HTML file to.

        equal_axes : bool, default True
            Set `xlim`, `ylim` to equal ranges such that the axes limits are
            equalised. Only has an effect if both `xlim` and `ylim` are `None`.
            If `False`, the default Plotly behaviour is used (i.e.
            automatically use min, max for each dimension).

        include_plotlyjs : True or "cdn", default True
            If `True`, embed the Plotly.JS library in the HTML file, allowing
            the graph to be shown offline, but adding 3 MB. If "cdn", the
            Plotly.JS library will be downloaded dynamically.

        Examples
        --------
        Add 10 random points to a `PlotlyGrapher2D` instance and save the
        figure as an HTML webpage:

        >>> fig = pept.visualisation.PlotlyGrapher2D()
        >>> fig.add_points(np.random.random((10, 3)))
        >>> fig.to_html("random_points.html")

        '''
        if equal_axes is True and self.xlim is None and self.ylim is None:
            self.equalise_axes()

        self._fig.write_html(
            filepath,
            include_plotlyjs = include_plotlyjs,
        )


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"xlim = {self.xlim}\n"
            f"ylim = {self.ylim}\n"
            f"zlim = {self.zlim}\n\n"
            f"fig = \n{self.fig}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL
        docstr = (
            "Class instance that inherits from `pept.visualisation."
            "PlotlyGrapher`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}\n\n"
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : parallel_map.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 03.02.2020


import  numpy               as      np
from    multiprocessing     import  Pool


def parallel_map_file(
    func,           # Called as func(data_chunk, chunk_number, *args, **kwargs)
    fname,          # File that will be supplied to numpy.loadtxt
    start,          # Start line
    end,            # End line
    chunksize,      # Number of lines per chunk
    *args,
    dtype = float,
    processes = None,
    callback = lambda x: None,
    error_callback = lambda x: None,
    **kwargs
):
    '''Utility for parallelising (read CSV chunk -> process chunk) workflows.

    This function reads individual chunks of data from a CSV-formatted file,
    then asynchronously sends them as numpy arrays to an arbitrary function
    `func` for processing. In effect, it reads a file in one main thread and
    processes it in separate threads.

    This is especially useful when dealing with very large files (like we do in
    PEPT...) that you'd like to process in batches, in parallel.

    Parameters
    ----------
    func : callable
        The function that will be called with each chunk of data, the chunk
        number, the other positional arguments `*args` and keyword arguments
        `**kwargs`: `func(data_chunk, chunk_number, *args, **kwargs)`.
        `data_chunk` is a numpy array returned by `numpy.loadtxt` and
        `chunk_number` is an int. `func` must be picklable for sending to
        other threads.
    fname : file, str, or pathlib.Path
        The file, filename, or generator that numpy.loadtxt will be supplied
        with.
    start : int
        The starting line number that the chunks will be read from.
    end : int
        The ending line number that the chunks will be read from. This is
        exclusive.
    chunksize : int
        The number of lines that will be read for each chunk.
    *args : additional positional arguments
        Additional positional arguments that will be supplied to `func`.
    dtype : type
        The data type of the numpy array that is returned by numpy.loadtxt. The
        default is `float`.
    processes : int
        The maximum number of threads that will be used for calling `func`. If
        left to the default `None`, then the number returned by
        `os.cpu_count()` will be used.
    callback : callable
        When the result from a `func` call becomes ready callback is applied to
        it, that is unless the call failed, in which case the error_callback is
        applied instead.
    error_callback : callable
        If the target function `func` fails, then the error_callback is called
        with the exception instance.
    **kwargs : additional keybord arguments
        Additional keyword arguments that will be supplied to `func`.

    Returns
    -------
    list
        A Python list of the `func` call returns. The results are not
        necessarily in order, though this can be verified by using the chunk
        number that is supplied to each call to `func`. If `func` does not
        return anything, it will simply be a list of `None`.

    Notes
    -----
    This function uses `numpy.loadtxt` to read chunks of data and
    `multiprocessing.Pool.apply_async` to call `func` asynchronously.

    As the calls to `func` happen in different threads, all the usual parallel
    processing issues apply. For example, `func` should not save data to the
    same file, as it will overwrite results from different threads and may
    become corrupt - however, there is a workaround for this particular case:
    because the chunk numbers are guaranteed to be unique, any data can be
    saved to a file whose name includes this chunk number, making it unique.

    Examples
    --------
    For a random file-like CSV data object:

    >>> import io
    >>> flike = io.StringIO("1,2,3\\n4,5,6\\n7,8,9")
    >>> def func(data, chunk_number):
    >>>     return (data, chunk_number)
    >>> results = parallel_map_file(func, flike, 0, 3, 1)
    >>> print(results)
    >>> [ ([1, 2, 3], 0), ([4, 5, 6], 1), ([7, 8, 9], 2) ]

    '''

    nchunks = int((end - start) / chunksize)

    with Pool(processes = processes) as pool:
        results = []
        for i in range(nchunks):
            data = np.loadtxt(
                fname,
                skiprows = start + i * chunksize,
                max_rows = chunksize,
                dtype = dtype
            )
            worker = pool.apply_async(
                func,
                (data, i, *args),
                kwargs,
                callback,
                error_callback
            )
            results.append(worker)

        results = [r.get() for r in results]

    return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : aggregate.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 10.06.2020


import numpy as np


def group_by_column(data_array, column_to_separate):
    '''Group the rows in a 2D `data_array` based on the unique values in a
    given `column_to_separate`, returning the groups as a list of numpy arrays.

    Parameters
    ----------
    data_array : (M, N) numpy.ndarray
        A generic 2D numpy array-like (will be converted using numpy.asarray).
    column_to_separate : int
        The column index in `data_array` from which the unique values will be
        used for grouping.

    Returns
    -------
    groups : list of numpy.ndarray
        A list whose elements are 2D numpy arrays - these are sub-arrays from
        `data_array` for which the entries in the column `column_to_separate`
        are the same.

    Raises
    ------
    ValueError
        If data_array does not have exactly 2 dimensions.

    Examples
    --------
    Separate a 6x3 numpy array based on the last column:

    >>> x = np.array([
    >>>     [1, 2, 1],
    >>>     [5, 3, 1],
    >>>     [1, 1, 2],
    >>>     [5, 2, 1],
    >>>     [2, 4, 2]
    >>> ])
    >>> x_sep = pept.utilities.group_by_column(x, -1)
    >>> x_sep
    >>> [array([[1, 2, 1],
    >>>         [5, 3, 1],
    >>>         [5, 2, 1]]),
    >>>  array([[1, 1, 2],
    >>>         [2, 4, 2]])]

    '''

    data_array = np.asarray(data_array)
    if data_array.ndim != 2:
        raise ValueError((
            "\n[ERROR]: `data_array` should have exactly 2 dimensions. "
            f"Received {data_array} with {data_array.ndim} dimensions.\n"
        ))

    data_col = data_array[:, column_to_separate]
    labels = np.unique(data_col)

    groups = [data_array[data_col == label] for label in labels]
    return groups



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : line_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.08.2019


import  time
import  pickle

import  numpy                   as      np

import  plotly.graph_objects    as      go
import  matplotlib.pyplot       as      plt

from    .iterable_samples       import  IterableSamples
import  pept


class LineData(IterableSamples):
    '''A class for PEPT LoR data iteration, manipulation and visualisation.

    Generally, PEPT Lines of Response (LoRs) are lines in 3D space, each
    defined by two points, regardless of the geometry of the scanner used. This
    class is used for the encapsulation of LoRs (or any lines!), efficiently
    yielding samples of `lines` of an adaptive `sample_size` and `overlap`.

    It is an abstraction over PET / PEPT scanner geometries and data formats,
    as once the raw LoRs (be they stored as binary, ASCII, etc.) are
    transformed into the common `LineData` format, any tracking, analysis or
    visualisation algorithm in the `pept` package can be used interchangeably.
    Moreover, it provides a stable, user-friendly interface for iterating over
    LoRs in *samples* - this is useful for tracking algorithms, as they
    generally take a few LoRs (a *sample*), produce a tracer position, then
    move to the next sample of LoRs, repeating the procedure. Using overlapping
    samples is also useful for improving the time resolution of the algorithms.

    This is the base class for LoR data; the subroutines for transforming other
    data formats into `LineData` can be found in `pept.scanners`. If you'd like
    to integrate another scanner geometry or raw data format into this package,
    you can check out the `pept.scanners.parallel_screens` module as an
    example. This usually only involves writing a single function by hand; then
    all attributes and methods from `LineData` will be available to your new
    data format. If you'd like to use `LineData` as the base for other
    algorithms, you can check out the `pept.tracking.peptml.cutpoints` module
    as an example; the `Cutpoints` class iterates the samples of LoRs in any
    `LineData` **in parallel**, using `concurrent.futures.ThreadPoolExecutor`.

    Attributes
    ----------
    lines : (N, M>=7) numpy.ndarray
        An (N, M>=7) numpy array that stores the PEPT LoRs as time and
        cartesian (3D) coordinates of two points defining a line, followed by
        any additional data. The data columns are then
        `[time, x1, y1, z1, x2, y2, z2, etc.]`.

    sample_size : int
        An `int` that defines the number of lines that should be returned when
        iterating over `lines`. The default is 0.

    overlap : int
        An `int` that defines the overlap between two consecutive samples that
        are returned when iterating over `lines`. An overlap of 0 implies
        consecutive samples, while an overlap of (`sample_size` - 1) implies
        incrementing the samples by one. A negative overlap means skipping
        values between samples. It is required to be smaller than
        `sample_size`. The default is 0.

    number_of_lines : int
        An `int` that corresponds to len(`lines`), or the number of LoRs
        stored by `lines`.

    number_of_samples : int
        An `int` that corresponds to the number of samples that can be accessed
        from the class. It takes `overlap` into consideration.

    Methods
    -------
    sample(n)
        Get sample number n (indexed from 0).

    to_csv(filepath)
        Write `lines` to a CSV file.

    save(filepath)
        Save a `LineData` instance as a binary `pickle` object.

    load(filepath)
        Load a saved / pickled `LineData` object from `filepath`.

    plot(sample_indices = ..., ax = None, alt_axes = False, colorbar_col = 0)
        Plot lines from selected samples using matplotlib.

    lines_trace(sample_indices = ..., width = 2, color = None, opacity = 0.6,\
                colorbar = True, colorbar_col = 0, colorbar_title = None)
        Get a Plotly trace for all the lines in selected samples.

    copy()
        Create a deep copy of an instance of this class, including a new inner
        numpy array `lines`.

    Raises
    ------
    ValueError
        If `overlap` >= `sample_size` unless `sample_size` is 0. Overlap
        must be smaller than `sample_size`. Note that it can also be negative.

    Notes
    -----
    The class saves `lines` as a **contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the number of rows or columns after instantiating the class.

    Examples
    --------
    Initialise a `LineData` instance containing 10 lines with a `sample_size`
    of 3.

    .. ipython::

        In [3]: import pept

        In [4]: import numpy as np

        In [5]: lines_raw = np.arange(70).reshape(10, 7)

        In [6]: print(lines_raw)
        [[ 0  1  2  3  4  5  6]
         [ 7  8  9 10 11 12 13]
         [14 15 16 17 18 19 20]
         [21 22 23 24 25 26 27]
         [28 29 30 31 32 33 34]
         [35 36 37 38 39 40 41]
         [42 43 44 45 46 47 48]
         [49 50 51 52 53 54 55]
         [56 57 58 59 60 61 62]
         [63 64 65 66 67 68 69]]

        In [7]: line_data = pept.LineData(lines_raw, sample_size = 3)

        In [8]: print(line_data)
        number_of_lines =   10
        sample_size =       3
        overlap =           0
        number_of_samples = 3
        lines =
        [[ 0.  1.  2.  3.  4.  5.  6.]
         [ 7.  8.  9. 10. 11. 12. 13.]
         [14. 15. 16. 17. 18. 19. 20.]
         [21. 22. 23. 24. 25. 26. 27.]
         [28. 29. 30. 31. 32. 33. 34.]
         [35. 36. 37. 38. 39. 40. 41.]
         [42. 43. 44. 45. 46. 47. 48.]
         [49. 50. 51. 52. 53. 54. 55.]
         [56. 57. 58. 59. 60. 61. 62.]
         [63. 64. 65. 66. 67. 68. 69.]]

    Access samples using subscript notation. Notice how the samples are
    consecutive, as `overlap` is 0 by default.

    .. ipython::

        In [9]: line_data[0]
        Out[9]:
        array([[ 0.,  1.,  2.,  3.,  4.,  5.,  6.],
               [ 7.,  8.,  9., 10., 11., 12., 13.],
               [14., 15., 16., 17., 18., 19., 20.]])

        In [10]: line_data[1]
        Out[10]:
        array([[21., 22., 23., 24., 25., 26., 27.],
               [28., 29., 30., 31., 32., 33., 34.],
               [35., 36., 37., 38., 39., 40., 41.]])

    Now set an overlap of 2; notice how the number of samples changes:

    .. ipython::

        In [11]: len(line_data)     # Number of samples
        Out[11]: 3

        In [12]: line_data.overlap = 2

        In [13]: len(line_data)
        Out[13]: 8

    Notice how rows are repeated from one sample to the next when accessing
    them, because `overlap` is now 2:

    .. ipython::

        In [14]: line_data[0]
        Out[14]:
        array([[ 0.,  1.,  2.,  3.,  4.,  5.,  6.],
               [ 7.,  8.,  9., 10., 11., 12., 13.],
               [14., 15., 16., 17., 18., 19., 20.]])

        In [15]: line_data[1]
        Out[15]:
        array([[ 7.,  8.,  9., 10., 11., 12., 13.],
               [14., 15., 16., 17., 18., 19., 20.],
               [21., 22., 23., 24., 25., 26., 27.]])

    Now change `sample_size` to 5 and notice again how the number of samples
    changes:

    .. ipython::

        In [16]: len(line_data)
        Out[16]: 8

        In [17]: line_data.sample_size = 5

        In [18]: len(line_data)
        Out[18]: 2

        In [19]: line_data[0]
        Out[19]:
        array([[ 0.,  1.,  2.,  3.,  4.,  5.,  6.],
               [ 7.,  8.,  9., 10., 11., 12., 13.],
               [14., 15., 16., 17., 18., 19., 20.],
               [21., 22., 23., 24., 25., 26., 27.],
               [28., 29., 30., 31., 32., 33., 34.]])

        In [20]: line_data[1]
        Out[20]:
        array([[21., 22., 23., 24., 25., 26., 27.],
               [28., 29., 30., 31., 32., 33., 34.],
               [35., 36., 37., 38., 39., 40., 41.],
               [42., 43., 44., 45., 46., 47., 48.],
               [49., 50., 51., 52., 53., 54., 55.]])

    Notice how the samples do not cover the whole input `lines_raw` array, as
    the last lines are omitted - think of the `sample_size` and `overlap`. They
    are still inside the inner `lines` attribute of `line_data` though:

    .. ipython::

        In [21]: line_data.lines
        Out[21]:
        array([[ 0.,  1.,  2.,  3.,  4.,  5.,  6.],
               [ 7.,  8.,  9., 10., 11., 12., 13.],
               [14., 15., 16., 17., 18., 19., 20.],
               [21., 22., 23., 24., 25., 26., 27.],
               [28., 29., 30., 31., 32., 33., 34.],
               [35., 36., 37., 38., 39., 40., 41.],
               [42., 43., 44., 45., 46., 47., 48.],
               [49., 50., 51., 52., 53., 54., 55.],
               [56., 57., 58., 59., 60., 61., 62.],
               [63., 64., 65., 66., 67., 68., 69.]])

    See Also
    --------
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    pept.tracking.peptml.Cutpoints : Compute cutpoints from `pept.LineData`.
    '''

    def __init__(
        self,
        lines,
        sample_size = 0,
        overlap = 0,
        verbose = False
    ):
        '''`LineData` class constructor.

        Parameters
        ----------
        lines : (N, M>=7) numpy.ndarray
            An (N, M>=7) numpy array that stores the PEPT LoRs (or any generic
            set of lines) as time and cartesian (3D) coordinates of two points
            defining each line, followed by any additional data. The data
            columns are then `[time, x1, y1, z1, x2, y2, z2, etc.]`.

        sample_size : int, default 0
            An `int` that defines the number of lines that should be returned
            when iterating over `lines`. A `sample_size` of 0 yields all the
            data as one single sample.

        overlap : int, default 0
            An `int` that defines the overlap between two consecutive samples
            that are returned when iterating over `lines`. An overlap of 0
            means consecutive samples, while an overlap of (`sample_size` - 1)
            means incrementing the samples by one. A negative overlap means
            skipping values between samples. An error is raised if `overlap` is
            larger than or equal to `sample_size`.

        verbose : bool, default False
            An option that enables printing the time taken for the
            initialisation of an instance of the class. Useful when reading
            large files (10gb files for PEPT data is not unheard of).

        Raises
        ------
        ValueError
            If `lines` has fewer than 7 columns.

        ValueError
            If `overlap` >= `sample_size` unless `sample_size` is 0. Overlap
            has to be smaller than `sample_size`. Note that it can also be
            negative.
        '''

        if verbose:
            start = time.time()

        # If `lines` is not C-contiguous, create a C-contiguous copy
        self._lines = np.asarray(lines, order = 'C', dtype = float)

        # Check that lines has at least 7 columns.
        if self._lines.ndim != 2 or self._lines.shape[1] < 7:
            raise ValueError((
                "\n[ERROR]: `lines` should have dimensions (M, N), where "
                f"N >= 7. Received {self._lines.shape}.\n"
            ))

        self._number_of_lines = len(self._lines)

        # Call the IterableSamples constructor to make the class iterable in
        # terms of samples with overlap.
        IterableSamples.__init__(self, sample_size, overlap)

        if verbose:
            end = time.time()
            print(f"Initialised LineData class in {end - start} s.\n")


    @property
    def lines(self):
        # The lines stored in the class.
        return self._lines


    @property
    def data_samples(self):
        # Implemented property for the IterableSamples parent class. See its
        # documentation for more details.
        return self._lines


    @property
    def data_length(self):
        # Implemented property for the IterableSamples parent class. See its
        # documentation for more details.
        return self._number_of_lines


    @property
    def number_of_lines(self):
        return self._number_of_lines


    def to_csv(self, filepath, delimiter = " "):
        '''Write `lines` to a CSV file.

        Write all LoRs stored in the class to a CSV file.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        delimiter : str, default " "
            The delimiter used to separate the values in the CSV file.

        '''

        np.savetxt(filepath, self._lines, delimiter = delimiter)


    def save(self, filepath):
        '''Save a `LineData` instance as a binary `pickle` object.

        Saves the full object state, including the inner `.lines` NumPy array,
        `sample_size`, etc. in a fast, portable binary format. Load back the
        object using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `LineData` instance, then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> lines.save("lines.pickle")

        >>> lines_reloaded = pept.LineData.load("lines.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `LineData` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.LineData
            The loaded `pept.LineData` instance.

        Examples
        --------
        Save a `LineData` instance, then load it back:

        >>> lines = pept.LineData([[1, 2, 3, 4, 5, 6, 7]])
        >>> lines.save("lines.pickle")

        >>> lines_reloaded = pept.LineData.load("lines.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def plot(
        self,
        sample_indices = ...,
        ax = None,
        alt_axes = False,
        colorbar_col = 0,
    ):
        '''Plot lines from selected samples using matplotlib.

        Returns matplotlib figure and axes objects containing all lines
        included in the samples selected by `sample_indices`.
        `sample_indices` may be a single sample index (e.g. 0), an iterable
        of indices (e.g. [1,5,6]), or an Ellipsis (`...`) for all samples.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of lines. An `int` signifies
            the sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all lines).

        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        alt_axes : bool, default False
            If `True`, plot using the alternative PEPT-style axes convention:
            z is horizontal, y points upwards. Because Matplotlib cannot swap
            axes, this is achieved by swapping the parameters in the plotting
            call (i.e. `plt.plot(x, y, z)` -> `plt.plot(z, x, y)`).

        colorbar_col : int, default -1
            The column in the data samples that will be used to color the
            lines. The default is -1 (the last column).

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Notes
        -----
        Plotting all lines is very computationally-expensive for matplotlib. It
        is recommended to only plot a couple of samples at a time, or use the
        faster `pept.visualisation.PlotlyGrapher`.

        Examples
        --------
        Plot the lines from sample 1 in a `LineData` instance:

        >>> lors = pept.LineData(...)
        >>> fig, ax = lors.plot(1)
        >>> fig.show()

        Plot the lines from samples 0, 1 and 2:

        >>> fig, ax = lors.plot([0, 1, 2])
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection = '3d')
        else:
            fig = plt.gcf()

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        lines, color = [], []
        # For each selected sample include all the lines' coordinates
        for n in sample_indices:
            # If an Ellipsis was received, then include all lines.
            if n is Ellipsis:
                sample = self.lines
            else:
                sample = self[n]

            for line in sample:
                lines.append(line[0:7])
                color.append(line[colorbar_col])

        color = np.array(color)

        # Scatter x, y, z [color]
        cmap = plt.cm.magma
        color_array = cmap(color / color.max())

        if alt_axes:
            for i, line in enumerate(lines):
                ax.plot(
                    [line[3], line[6]],
                    [line[1], line[4]],
                    [line[2], line[5]],
                    c = color_array[i],
                    alpha = 0.8
                )

            ax.set_xlabel("z (mm)")
            ax.set_ylabel("x (mm)")
            ax.set_zlabel("y (mm)")

        else:
            for i, line in enumerate(lines):
                ax.plot(
                    [line[1], line[4]],
                    [line[2], line[5]],
                    [line[3], line[6]],
                    c = color_array[i],
                    alpha = 0.8
                )

            ax.set_xlabel("x (mm)")
            ax.set_ylabel("y (mm)")
            ax.set_zlabel("z (mm)")

        return fig, ax


    def lines_trace(
        self,
        sample_indices = ...,
        width = 2.0,
        color = None,
        opacity = 0.6,
        colorbar = True,
        colorbar_col = 0,
        colorscale = "Magma",
        colorbar_title = None
    ):
        '''Get a Plotly trace for all the lines in selected samples.

        Creates a `plotly.graph_objects.Scatter3d` object for all the lines
        included in the samples selected by `sample_indices`. `sample_indices`
        may be a single sample index (e.g. 0), an iterable of indices (e.g.
        [1,5,6]) or an Ellipsis (`...`) for all samples.

        Can then be passed to the `plotly.graph_objects.figure.add_trace`
        function or a `PlotlyGrapher` instance using the `add_trace` method.

        Parameters
        ----------
        sample_indices : int or iterable or Ellipsis, default Ellipsis
            The index or indices of the samples of LoRs. An `int` signifies the
            sample index, an iterable (list-like) signifies multiple sample
            indices, while an Ellipsis (`...`) signifies all samples. The
            default is `...` (all lines).

        width : float, default 2.0
            The width of the lines.

        color : str or list-like, optional
            Can be a single color (e.g. "black", "rgb(122, 15, 241)") or a
            colorbar list. Overrides `colorbar` if set. For more information,
            check the Plotly documentation. The default is None.

        opacity : float, default 0.6
            The opacity of the lines, where 0 is transparent and 1 is fully
            opaque.

        colorbar : bool, default True
            If set to True, will color-code the data in the sample column
            `colorbar_col`. Is overridden if `color` is set. The default is
            True, so that every line has a different color.

        colorbar_col : int, default 0
            The column in the data samples that will be used to color the
            points. Only has an effect if `colorbar` is set to True. The
            default is 0 (the first column - time).

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the `colorbar_col` column in the
            input data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        colorbar_title : str, optional
            If set, the colorbar will have this title above.

        Returns
        -------
        plotly.graph_objs.Scatter3d
            A Plotly trace of the LoRs.

        Examples
        --------
        Use `PlotlyGrapher` (a user-friendly wrapper around the `plotly`
        library for PEPT-oriented data) to plot the lines from sample 1 in a
        `LineData` instance:

        >>> lors = pept.LineData(...)
        >>> grapher = pept.visualisation.PlotlyGrapher()
        >>> trace = lors.lines_trace(1)
        >>> grapher.add_trace(trace)
        >>> grapher.show()

        Use `plotly.graph_objs` to plot the lines from samples 0, 1 and 2:

        >>> import plotly.graph_objs as go
        >>> fig = go.Figure()
        >>> fig.add_trace(lors.lines_trace([0, 1, 2]))
        >>> fig.show()

        '''

        # Check if sample_indices is an iterable collection (list-like)
        # otherwise just "iterate" over the single number or Ellipsis.
        if not hasattr(sample_indices, "__iter__"):
            sample_indices = [sample_indices]

        marker = dict(
            width = width,
            color = color,
        )

        if colorbar:
            if color is None:
                marker['color'] = []

            marker.update(colorscale = colorscale)
            if colorbar_title is not None:
                marker.update(colorbar = dict(title = colorbar_title))

        coords_x = []
        coords_y = []
        coords_z = []

        # For each selected sample include all the lines' coordinates
        for n in sample_indices:
            # If an Ellipsis was received, then include all lines.
            if n is Ellipsis:
                sample = self.lines
            else:
                sample = self[n]

            for line in sample:
                coords_x.extend([line[1], line[4], None])
                coords_y.extend([line[2], line[5], None])
                coords_z.extend([line[3], line[6], None])

                if colorbar and color is None:
                    marker['color'].extend(3 * [line[colorbar_col]])

        trace = go.Scatter3d(
            x = coords_x,
            y = coords_y,
            z = coords_z,
            mode = 'lines',
            opacity = opacity,
            line = marker
        )

        return trace


    def copy(self):
        '''Create a deep copy of an instance of this class, including a new
        inner numpy array `lines`.

        Returns
        -------
        pept.LineData
            A new instance of the `pept.LineData` class with the same
            attributes as this instance, deep-copied.
        '''

        return pept.LineData(
            self._lines.copy(order = "C"),
            sample_size = self._sample_size,
            overlap = self._overlap,
            verbose = False
        )


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"number_of_lines =   {self.number_of_lines}\n\n"
            f"sample_size =       {self._sample_size}\n"
            f"overlap =           {self._overlap}\n"
            f"number_of_samples = {self.number_of_samples}\n\n"
            f"lines = \n{self._lines}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPR
        docstr = (
            "Class instance that inherits from `pept.LineData`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}\n\n"
            "Particular Cases\n----------------\n"
            " > If sample_size == 0, all `lines` are returned as a "
            "single sample.\n"
            " > If overlap >= sample_size, an error is raised.\n"
            " > If overlap < 0, lines are skipped between samples."
        )

        return docstr



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : distance_matrix_reachable.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 10.06.2020


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np
from scipy.sparse import csr_matrix

from libc.float cimport DBL_MIN
from libc.math cimport sqrt
cimport numpy as np


cpdef distance_matrix_reachable(
    double[:, :] pts,           # Array of points, cols = [t, x, y, z, ...]
    int points_window
):
    '''Compute the distance matrix from a time-sorted array of points `pts`
    based on a sliding `points_window`.

    ::

        Function signature:
            distance_matrix_reachable(
                double[:, :] pts,   # Array of points, cols = [t, x, y, z, ...]
                int points_window
            )

    The distance between the points (pts[i], pts[j]) is stored in the distance
    matrix at indices (i, j), making it upper-triangular.

    The distance matrix is created and returned using SciPy's sparse CSR
    matrix format. This saves a lot in terms of memory usage, especially for
    time-series data such as moving tracers, as not all points can be
    connected. This format is also closer to the mathematical formulation of
    "reachable points" in terms of undirected (incomplete) graphs - namely
    storing edges as a list of pairs of vertices.

    This is a low-level Cython function that does not do any checks on the
    input data - it is meant to be used in other modules / libraries; in
    particular, the `pept.tracking.trajectory_separation` module.

    Parameters
    ----------
    pts : (M, N>=4) numpy.ndarray
        The points from multiple trajectories. Each row in `pts` will
        have a timestamp and the 3 spatial coordinates, such that the data
        columns are [time, x_coord, y_coord, z_coord]. Note that `point_data`
        can have more data columns and they will simply be ignored.
    points_window : int
        Two points are "reachable" (i.e. they can be connected) if and only if
        they are within `points_window` in the time-sorted input `pts`. As the
        points from different trajectories are intertwined (e.g. for two
        tracers A and B, the `pts` array might have two entries for A,
        followed by three entries for B, then one entry for A, etc.), this
        should optimally be the largest number of points in the input array
        between two consecutive points on the same trajectory. If
        `pts` is too small, all points in the dataset will be unreachable.
        Naturally, a larger `time_window` correponds to more pairs needing to
        be checked (and the function will take a longer to complete).

    Returns
    -------
    distance_matrix : CSR <NxN sparse matrix of type '<class 'numpy.float64'>
        A SciPy sparse matrix in the CSR format, containing the distances
        between every pair of reachable points in `pts`.

    Notes
    -----
    In order for the `points_window` to act as a sliding window, in effect only
    connecting points which are around the same timeframe, the points should be
    sorted based on thetime column (the first row) in `pts`. This should be
    done *prior* to calling this function.
    '''

    # Use Py_ssize_t as we will access C arrays (memoryviews on numpy arrays).
    # That is the "proper" type of a C array pointer / index.
    cdef Py_ssize_t n = pts.shape[0]            # Total number of points
    cdef Py_ssize_t p = min(n, points_window)

    # Calculate sparse distance matrix between reachable points. The number of
    # points we need to check `ndists` is given by the formula below.
    cdef Py_ssize_t ndists = (p + 1) * (n - p) + p * (p + 1) // 2 - n

    # Pre-allocate the arrays for creating the sparse distance matrix. In the
    # sparse matrix, every data point `dists` has an associated row in `rows`
    # column in `cols`.
    cdef np.ndarray[double, ndim = 1] dists_arr = np.zeros(ndists, dtype =
                                                           np.float64)
    cdef np.ndarray[double, ndim = 1] rows_arr = np.zeros(ndists, dtype =
                                                          np.float64)
    cdef np.ndarray[double, ndim = 1] cols_arr = np.zeros(ndists, dtype =
                                                          np.float64)

    # We'll work with memoryviews on the above arrays:
    cdef double[:] dists = dists_arr
    cdef double[:] rows = rows_arr
    cdef double[:] cols = cols_arr

    # Calculate the distances between reachable points.
    cdef Py_ssize_t ie = 0  # distance index
    cdef Py_ssize_t i, j    # iterators
    cdef double dist        # distance between two points

    with nogil:
        for i in range(n - 1):
            for j in range(i + 1, min(i + p, n - 1) + 1):
                # Euclidean distance between points i, j in `pts`
                # dist = np.linalg.norm(pts[i, 1:4] - pts[j, 1:4])
                dist = sqrt(
                    (pts[j, 1] - pts[i, 1]) ** 2 +
                    (pts[j, 2] - pts[i, 2]) ** 2 +
                    (pts[j, 3] - pts[i, 3]) ** 2
                )

                # Fix bug (or feature?) of scipy's minimum_spanning_tree where
                # duplicate points (i.e. dist == 0.0) are ommitted from the
                # MST vertices.
                dists[ie] = dist if dist != 0.0 else DBL_MIN
                rows[ie] = i
                cols[ie] = j
                ie = ie + 1

    # Create the distance matrix from the found points.
    distance_matrix = csr_matrix(
        (dists, (rows, cols)),
        shape = (n, n)
    )

    return distance_matrix



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : pixel_data.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 07.01.2020


import  pickle
import  time
import  textwrap

import  numpy                   as      np

import  plotly.graph_objects    as      go
import  matplotlib.pyplot       as      plt

from    pept.utilities.traverse import  traverse2d




class Pixels(np.ndarray):
    '''A class that manages a 2D pixel space, including tools for pixel
    traversal of lines, manipulation and visualisation.

    This class can be instantiated in a couple of ways:

    1. The constructor receives a pre-defined pixel space (i.e. a 2D numpy
       array), along with the space boundaries `xlim` and `ylim`.

    2. The `from_lines` method receives a sample of 2D lines (i.e. a 2D numpy
       array), each defined by two points, creating a pixel space and
       traversing / pixellising the lines.

    3. The `empty` method creates a pixel space filled with zeros.

    This subclasses the `numpy.ndarray` class, so any `Pixels` object acts
    exactly like a 2D numpy array. All numpy methods and operations are valid
    on `Pixels` (e.g. add 1 to all pixels with `pixels += 1`).

    It is possible to add multiple samples of lines to the same pixel space
    using the `add_lines` method.

    Attributes
    ----------
    pixels: (M, N) numpy.ndarray
        The 2D numpy array containing the number of lines that pass through
        each pixel. They are stored as `float`s. This class assumes a uniform
        grid of pixels - that is, the pixel size in each dimension is constant,
        but can vary from one dimension to another. The number of pixels in
        each dimension is defined by `number_of_pixels`.

    number_of_pixels: 2-tuple
        A 2-tuple corresponding to the shape of `pixels`.

    pixel_size: (2,) numpy.ndarray
        The lengths of a pixel in the x- and y-dimensions, respectively.

    xlim: (2,) numpy.ndarray
        The lower and upper boundaries of the pixellised volume in the
        x-dimension, formatted as [x_min, x_max].

    ylim: (2,) numpy.ndarray
        The lower and upper boundaries of the pixellised volume in the
        y-dimension, formatted as [y_min, y_max].

    pixel_grids: list[numpy.ndarray]
        A list containing the pixel gridlines in the x- and y-dimensions.
        Each dimension's gridlines are stored as a numpy of the pixel
        delimitations, such that it has length (M + 1), where M is the number
        of pixels in a given dimension.

    Methods
    -------
    save(filepath)
        Save a `Pixels` instance as a binary `pickle` object.

    load(filepath)
        Load a saved / pickled `Pixels` object from `filepath`.

    from_lines(lines, number_of_pixels, xlim = None, ylim = None, \
               verbose = True)
        Create a pixel space and traverse / pixellise a given sample of
        `lines`.

    empty(number_of_pixels, xlim, ylim, verbose = False)
        Create an empty pixel space for the 2D rectangle bounded by `xlim` and
        `ylim`.

    get_cutoff(p1, p2)
        Return a numpy array containing the minimum and maximum value found
        across the two input arrays.

    add_lines(lines, verbose = False)
        Pixellise a sample of lines, adding 1 to each pixel traversed, for
        each line in the sample.

    cube_trace(index, color = None, opacity = 0.4, colorbar = True,\
               colorscale = "magma")
        Get the Plotly `Mesh3d` trace for a single pixel at `index`.

    cubes_traces(condition = lambda pixels: pixels > 0, color = None,\
                 opacity = 0.4, colorbar = True, colorscale = "magma")
        Get a list of Plotly `Mesh3d` traces for all pixel selected by the
        `condition` filtering function.

    pixels_trace(condition = lambda pixels: pixels > 0, size = 4,\
                 color = None, opacity = 0.4, colorbar = True,\
                 colorscale = "Magma", colorbar_title = None)
        Create and return a trace for all the pixels in this class, with
        possible filtering.

    heatmap_trace(ix = None, iy = None, iz = None, width = 0,\
                  colorscale = "Magma", transpose = True)
        Create and return a Plotly `Heatmap` trace of a 2D slice through the
        voxels.

    Notes
    -----
    The traversed lines do not need to be fully bounded by the pixel space.
    Their intersection is automatically computed.

    The class saves `pixels` as a **contiguous** numpy array for efficient
    access in C / Cython functions. The inner data can be mutated, but do not
    change the shape of the array after instantiating the class.

    Examples
    --------
    This class is most often instantiated from a sample of lines to pixellise:

    >>> import pept
    >>> import numpy as np

    >>> lines = np.arange(70).reshape(10, 7)

    >>> number_of_pixels = [3, 4]
    >>> pixels = pept.Pixels.from_lines(lines, number_of_pixels)
    >>> Initialised Pixels class in 0.0006861686706542969 s.

    >>> print(pixels)
    >>> pixels:
    >>> [[[2. 1. 0. 0. 0.]
    >>>   [0. 2. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]]

    >>>  [[0. 0. 0. 0. 0.]
    >>>   [0. 1. 1. 0. 0.]
    >>>   [0. 0. 1. 1. 0.]
    >>>   [0. 0. 0. 0. 0.]]

    >>>  [[0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 0. 0.]
    >>>   [0. 0. 0. 2. 0.]
    >>>   [0. 0. 0. 1. 2.]]]

    >>> number_of_pixels =    (3, 4, 5)
    >>> pixel_size =          [22.  16.5 13.2]

    >>> xlim =                [ 1. 67.]
    >>> ylim =                [ 2. 68.]
    >>> zlim =                [ 3. 69.]

    >>> pixel_grids:
    >>> [array([ 1., 23., 45., 67.]),
    >>>  array([ 2. , 18.5, 35. , 51.5, 68. ]),
    >>>  array([ 3. , 16.2, 29.4, 42.6, 55.8, 69. ])]

    Note that it is important to define the `number_of_pixels`.

    See Also
    --------
    pept.LineData : Encapsulate lines for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.
    '''

    def __new__(
        cls,
        pixels_array,
        xlim,
        ylim,
    ):
        '''`Pixels` class constructor.

        Parameters
        ----------
        pixels_array: 3D numpy.ndarray
            A 3D numpy array, corresponding to a pre-defined pixel space.

        xlim: (2,) numpy.ndarray
            The lower and upper boundaries of the pixellised volume in the
            x-dimension, formatted as [x_min, x_max].

        ylim: (2,) numpy.ndarray
            The lower and upper boundaries of the pixellised volume in the
            y-dimension, formatted as [y_min, y_max].

        Raises
        ------
        ValueError
            If `pixels_array` does not have exactly 3 dimensions or if
            `xlim` or `ylim` do not have exactly 2 values each.

        '''

        # Type-checking inputs
        pixels_array = np.asarray(
            pixels_array,
            order = "C",
            dtype = float
        )

        if pixels_array.ndim != 2:
            raise ValueError(textwrap.fill((
                "The input `pixels_array` must contain an array-like with "
                "exactly three dimensions (i.e. pre-made pixels array). "
                f"Received an array with {pixels_array.ndim} dimensions. "
                "Note: if you would like to create pixels from a sample of"
                "lines, use the `Pixels.from_lines` method. "
            )))

        xlim = np.asarray(xlim, dtype = float)

        if xlim.ndim != 1 or len(xlim) != 2:
            raise ValueError(textwrap.fill((
                "The input `xlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the x-dimension. "
                f"Received parameter with shape {xlim.shape}."
            )))

        ylim = np.asarray(ylim, dtype = float)

        if ylim.ndim != 1 or len(ylim) != 2:
            raise ValueError(textwrap.fill((
                "The input `ylim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the y-dimension. "
                f"Received parameter with shape {ylim.shape}."
            )))

        # Setting class attributes
        pixels = pixels_array.view(cls)
        pixels._number_of_pixels = pixels.shape

        pixels._xlim = xlim
        pixels._ylim = ylim

        pixels._pixel_size = np.array([
            (pixels._xlim[1] - pixels._xlim[0]) / pixels._number_of_pixels[0],
            (pixels._ylim[1] - pixels._ylim[0]) / pixels._number_of_pixels[1],
        ])

        pixels._pixel_grids = tuple([
            np.linspace(lim[0], lim[1], pixels._number_of_pixels[i] + 1)
            for i, lim in enumerate((pixels._xlim, pixels._ylim))
        ])

        return pixels


    def __array_finalize__(self, pixels):
        # Required method for numpy subclassing
        if pixels is None:
            return

        self._number_of_pixels = getattr(pixels, "_number_of_pixels", None)
        self._pixel_size = getattr(pixels, "_pixel_size", None)

        self._xlim = getattr(pixels, "_xlim", None)
        self._ylim = getattr(pixels, "_ylim", None)
        self._zlim = getattr(pixels, "_zlim", None)

        self._pixel_grids = getattr(pixels, "_pixel_grids", None)


    def __reduce__(self):
        # __reduce__ and __setstate__ ensure correct pickling behaviour. See
        # https://stackoverflow.com/questions/26598109/preserve-custom-
        # attributes-when-pickling-subclass-of-numpy-array

        # Get the parent's __reduce__ tuple
        pickled_state = super(Pixels, self).__reduce__()

        # Create our own tuple to pass to __setstate__
        new_state = pickled_state[2] + (
            self._number_of_pixels,
            self._xlim,
            self._ylim,
            self._pixel_size,
            self._pixel_grids,
        )

        # Return a tuple that replaces the parent's __setstate__ tuple with
        # our own
        return (pickled_state[0], pickled_state[1], new_state)


    def __setstate__(self, state):
        # __reduce__ and __setstate__ ensure correct pickling behaviour
        # https://stackoverflow.com/questions/26598109/preserve-custom-
        # attributes-when-pickling-subclass-of-numpy-array

        # Set the class attributes
        self._pixel_grids = state[-1]
        self._pixel_size = state[-2]
        self._ylim = state[-3]
        self._xlim = state[-4]
        self._number_of_pixels = state[-5]

        # Call the parent's __setstate__ with the other tuple elements.
        super(Pixels, self).__setstate__(state[0:-5])


    @property
    def pixels(self):
        return self.__array__()


    @property
    def number_of_pixels(self):
        return self._number_of_pixels


    @property
    def xlim(self):
        return self._xlim


    @property
    def ylim(self):
        return self._ylim


    @property
    def pixel_size(self):
        return self._pixel_size


    @property
    def pixel_grids(self):
        return self._pixel_grids


    @staticmethod
    def from_lines(
        lines,
        number_of_pixels,
        xlim = None,
        ylim = None,
        verbose = True,
    ):
        '''Create a pixel space and traverse / pixellise a given sample of
        `lines`.

        The `number_of_pixels` in each dimension must be defined. If the
        pixel space boundaries `xlim` or `ylim` are not defined, they
        are inferred as the boundaries of the `lines`.

        Parameters
        ----------
        lines : (M, N>=5) numpy.ndarray
            The lines that will be pixellised, each defined by a timestamp and
            two 2D points, so that the data columns are [time, x1, y1, x2, y2].
            Note that extra columns are ignored.

        number_of_pixels : (2,) list[int]
            The number of pixels in the x- and y-dimensions, respectively.

        xlim : (2,) list[float], optional
            The lower and upper boundaries of the pixellised volume in the
            x-dimension, formatted as [x_min, x_max]. If undefined, it is
            inferred from the boundaries of `lines`.

        ylim : (2,) list[float], optional
            The lower and upper boundaries of the pixellised volume in the
            y-dimension, formatted as [y_min, y_max]. If undefined, it is
            inferred from the boundaries of `lines`.

        Returns
        -------
        pept.Pixels
            A new `Pixels` object with the pixels through which the lines were
            traversed.

        Raises
        ------
        ValueError
            If the input `lines` does not have the shape (M, N>=5). If the
            `number_of_pixels` is not a 1D list with exactly 2 elements, or
            any dimension has fewer than 2 pixels.

        '''
        if verbose:
            start = time.time()

        # Type-checking inputs
        lines = np.asarray(lines, order = "C", dtype = float)

        if lines.ndim != 2 or lines.shape[1] < 5:
            raise ValueError(textwrap.fill((
                "The input `lines` must be a 2D numpy array containing lines "
                "defined by a timestamp and two 2D points, with every row "
                "formatted as [t, x1, y1, x2, y2]. The `lines` must then have "
                f"shape (M, 5). Received array with shape {lines.shape}."
            )))

        number_of_pixels = np.asarray(
            number_of_pixels,
            order = "C",
            dtype = int
        )

        if number_of_pixels.ndim != 1 or len(number_of_pixels) != 2:
            raise ValueError(textwrap.fill((
                "The input `number_of_pixels` must be a list-like "
                "with exactly two values, corresponding to the "
                "number of pixels in the x- and y-dimension. "
                f"Received parameter with shape {number_of_pixels.shape}."
            )))

        if (number_of_pixels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_pixels` must set at least two "
                "pixels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_pixels}`."
            )))

        if xlim is None:
            xlim = Pixels.get_cutoff(lines[:, 1], lines[:, 3])
        else:
            xlim = np.asarray(xlim, dtype = float)

            if xlim.ndim != 1 or len(xlim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `xlim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the pixel space in the x-dimension. "
                    f"Received parameter with shape {xlim.shape}."
                )))

        if ylim is None:
            ylim = Pixels.get_cutoff(lines[:, 1], lines[:, 3])
        else:
            ylim = np.asarray(ylim, dtype = float)

            if ylim.ndim != 1 or len(ylim) != 2:
                raise ValueError(textwrap.fill((
                    "The input `ylim` parameter must be a list with exactly "
                    "two values, corresponding to the minimum and maximum "
                    "coordinates of the pixel space in the y-dimension. "
                    f"Received parameter with shape {ylim.shape}."
                )))

        pixels_array = np.zeros(tuple(number_of_pixels))
        pixels = Pixels(
            pixels_array,
            xlim = xlim,
            ylim = ylim,
        )

        pixels.add_lines(lines, verbose = False)

        if verbose:
            end = time.time()
            print((
                f"Initialised Pixels class in {end - start} s."
            ))

        return pixels


    @staticmethod
    def empty(number_of_pixels, xlim, ylim):
        '''Create an empty pixel space for the 3D cube bounded by `xlim` and
        `ylim`.

        Parameters
        ----------
        number_of_pixels: (2,) numpy.ndarray
            A list-like containing the number of pixels to be created in the
            x- and y-dimension, respectively.

        xlim: (2,) numpy.ndarray
            The lower and upper boundaries of the pixellised volume in the
            x-dimension, formatted as [x_min, x_max].

        ylim: (2,) numpy.ndarray
            The lower and upper boundaries of the pixellised volume in the
            y-dimension, formatted as [y_min, y_max].
            Time the pixellisation step and print it to the terminal.

        Raises
        ------
        ValueError
            If `number_of_pixels` does not have exactly 2 values, or it has
            values smaller than 2. If `xlim` or `ylim` do not have exactly 2
            values each.

        '''

        number_of_pixels = np.asarray(
            number_of_pixels,
            order = "C",
            dtype = int
        )

        if number_of_pixels.ndim != 1 or len(number_of_pixels) != 2:
            raise ValueError(textwrap.fill((
                "The input `number_of_pixels` must be a list-like "
                "with exactly three values, corresponding to the "
                "number of pixels in the x- and y-dimension. "
                f"Received parameter with shape {number_of_pixels.shape}."
            )))

        if (number_of_pixels < 2).any():
            raise ValueError(textwrap.fill((
                "The input `number_of_pixels` must set at least two "
                "pixels in each dimension (i.e. all elements in "
                "`number_of_elements` must be larger or equal to two). "
                f"Received `{number_of_pixels}`."
            )))

        number_of_pixels = tuple(number_of_pixels)
        empty_pixels = np.zeros(number_of_pixels)

        return Pixels(
            empty_pixels,
            xlim = xlim,
            ylim = ylim,
        )


    @staticmethod
    def get_cutoff(p1, p2):
        '''Return a numpy array containing the minimum and maximum value found
        across the two input arrays.

        Parameters
        ----------
        p1 : (N,) numpy.ndarray
            The first 1D numpy array.

        p2 : (N,) numpy.ndarray
            The second 1D numpy array.

        Returns
        -------
        (2,) numpy.ndarray
            The minimum and maximum value found across `p1` and `p2`.

        Notes
        -----
        The input parameters *must* be numpy arrays, otherwise an error will
        be raised.

        '''

        return np.array([
            min(p1.min(), p2.min()),
            max(p1.max(), p2.max()),
        ])


    def save(self, filepath):
        '''Save a `Pixels` instance as a binary `pickle` object.

        Saves the full object state, including the inner `.pixels` NumPy array,
        `xlim`, etc. in a fast, portable binary format. Load back the object
        using the `load` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Examples
        --------
        Save a `Pixels` instance, then load it back:

        >>> pixels = pept.Pixels.empty((640, 480), [0, 20], [0, 10])
        >>> pixels.save("pixels.pickle")

        >>> pixels_reloaded = pept.Pixels.load("pixels.pickle")

        '''
        with open(filepath, "wb") as f:
            pickle.dump(self, f)


    @staticmethod
    def load(filepath):
        '''Load a saved / pickled `Pixels` object from `filepath`.

        Most often the full object state was saved using the `.save` method.

        Parameters
        ----------
        filepath : filename or file handle
            If filepath is a path (rather than file handle), it is relative
            to where python is called.

        Returns
        -------
        pept.Pixels
            The loaded `pept.Pixels` instance.

        Examples
        --------
        Save a `Pixels` instance, then load it back:

        >>> pixels = pept.Pixels.empty((640, 480), [0, 20], [0, 10])
        >>> pixels.save("pixels.pickle")

        >>> pixels_reloaded = pept.Pixels.load("pixels.pickle")

        '''
        with open(filepath, "rb") as f:
            obj = pickle.load(f)

        return obj


    def add_lines(self, lines, verbose = False):
        '''Pixellise a sample of lines, adding 1 to each pixel traversed, for
        each line in the sample.

        Parameters
        ----------
        lines : (M, N >= 5) numpy.ndarray
            The sample of 2D lines to pixellise. Each line is defined as a
            timestamp followed by two 2D points, such that the data columns are
            `[time, x1, y1, x2, y2, ...]`. Note that there can be extra data
            columns which will be ignored.

        verbose : bool, default False
            Time the pixel traversal and print it to the terminal.

        Raises
        ------
        ValueError
            If `lines` has fewer than 5 columns.

        '''

        lines = np.asarray(lines, order = "C", dtype = float)
        if lines.ndim != 2 or lines.shape[1] < 5:
            raise ValueError(textwrap.fill((
                "The input `lines` must be a 2D array of lines, where each "
                "line (i.e. row) is defined by a timestamp and two 2D points, "
                "so the data columns are [time, x1, y1, x2, y2]. "
                f"Received array of shape {lines.shape}."
            )))

        if verbose:
            start = time.time()

        traverse2d(
            self.pixels,
            lines,
            self._pixel_grids[0],
            self._pixel_grids[1],
        )

        if verbose:
            end = time.time()
            print(f"Traversing {len(lines)} lines took {end - start} s.")


    def pixels_trace(
        self,
        condition = lambda pixels: pixels > 0,
        opacity = 0.9,
        colorscale = "Magma",
    ):
        '''Create and return a trace with all the pixels in this class, with
        possible filtering.

        Creates a `plotly.graph_objects.Surface` object for the centres of
        all pixels encapsulated in a `pept.Pixels` instance, colour-coding the
        pixel value.

        The `condition` parameter is a filtering function that should return
        a boolean mask (i.e. it is the result of a condition evaluation). For
        example `lambda x: x > 0` selects all pixels that have a value larger
        than 0.

        Parameters
        ----------
        condition : function, default `lambda pixels: pixels > 0`
            The filtering function applied to the pixel data before plotting
            it. It should return a boolean mask (a numpy array of the same
            shape, filled with True and False), selecting all pixels that
            should be plotted. The default, `lambda x: x > 0` selects all
            pixels which have a value larger than 0.

        opacity : float, default 0.4
            The opacity of the surface, where 0 is transparent and 1 is fully
            opaque.

        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the voxel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        Examples
        --------
        Pixellise an array of lines and add them to a `PlotlyGrapher` instance:

        >>> grapher = PlotlyGrapher()
        >>> lines = np.array(...)                   # shape (N, M >= 7)
        >>> lines2d = lines[:, [0, 1, 2, 4, 5]]     # select x, y of lines
        >>> number_of_pixels = [10, 10]
        >>> pixels = pept.Pixels.from_lines(lines2d, number_of_pixels)
        >>> grapher.add_lines(lines)
        >>> grapher.add_trace(pixels.pixels_trace())
        >>> grapher.show()

        '''

        filtered = self.copy()
        filtered[~condition(self)] = 0.

        # Compute the pixel centres
        x = self.pixel_grids[0]
        x = (x[1:] + x[:-1]) / 2

        y = self.pixel_grids[1]
        y = (y[1:] + y[:-1]) / 2

        trace = go.Surface(
            x = x,
            y = y,
            z = filtered,
            opacity = opacity,
            colorscale = colorscale,
        )

        return trace


    def heatmap_trace(
        self,
        colorscale = "Magma",
        transpose = True,
        xgap = 0.,
        ygap = 0.,
    ):
        '''Create and return a Plotly `Heatmap` trace of the pixels.

        Parameters
        ----------
        colorscale : str, default "Magma"
            The Plotly scheme for color-coding the pixel values in the input
            data. Typical ones include "Cividis", "Viridis" and "Magma".
            A full list is given at `plotly.com/python/builtin-colorscales/`.
            Only has an effect if `colorbar = True` and `color` is not set.

        transpose : bool, default True
            Transpose the heatmap (i.e. flip it across its diagonal).

        Examples
        --------
        Pixellise an array of lines and add them to a `PlotlyGrapher2D`
        instance:

        >>> lines = np.array(...)                   # shape (N, M >= 7)
        >>> lines2d = lines[:, [0, 1, 2, 4, 5]]     # select x, y of lines
        >>> number_of_pixels = [10, 10]
        >>> pixels = pept.Pixels.from_lines(lines2d, number_of_pixels)

        >>> grapher = pept.visualisation.PlotlyGrapher2D()
        >>> grapher.add_pixels(pixels)
        >>> grapher.show()

        Or add them directly to a raw `plotly.graph_objs` figure:

        >>> import plotly.graph_objs as go
        >>> fig = go.Figure()
        >>> fig.add_trace(pixels.heatmap_trace())
        >>> fig.show()

        '''

        # Compute the pixel centres
        x = self.pixel_grids[0]
        x = (x[1:] + x[:-1]) / 2

        y = self.pixel_grids[1]
        y = (y[1:] + y[:-1]) / 2

        heatmap = dict(
            x = x,
            y = y,
            z = self,
            colorscale = colorscale,
            transpose = transpose,
            xgap = xgap,
            ygap = ygap,
        )

        return go.Heatmap(heatmap)


    def plot(self, ax = None):
        '''Plot pixels as a heatmap using Matplotlib.

        Returns matplotlib figure and axes objects containing the pixel values
        colour-coded in a Matplotlib image (i.e. heatmap).

        Parameters
        ----------
        ax : mpl_toolkits.mplot3D.Axes3D object, optional
            The 3D matplotlib-based axis for plotting. If undefined, new
            Matplotlib figure and axis objects are created.

        Returns
        -------
        fig, ax : matplotlib figure and axes objects

        Examples
        --------
        Pixellise an array of lines and plot them with Matplotlib:

        >>> lines = np.array(...)                   # shape (N, M >= 7)
        >>> lines2d = lines[:, [0, 1, 2, 4, 5]]     # select x, y of lines
        >>> number_of_pixels = [10, 10]
        >>> pixels = pept.Pixels.from_lines(lines2d, number_of_pixels)

        >>> fig, ax = pixels.plot()
        >>> fig.show()

        '''

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            fig = plt.gcf()

        # Plot the values in pixels (this class is a numpy array subclass)
        ax.imshow(self)

        # Compute the pixel centres and set them in the Matplotlib image
        x = self.pixel_grids[0]
        x = (x[1:] + x[:-1]) / 2

        y = self.pixel_grids[1]
        y = (y[1:] + y[:-1]) / 2

        # Matplotlib shows numbers in a long format ("102.000032411"), so round
        # them to two decimals before plotting
        ax.set_xticklabels(np.round(x, 2))
        ax.set_yticklabels(np.round(y, 2))

        ax.set_xlabel("x (mm)")
        ax.set_ylabel("y (mm)")

        return fig, ax


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"{self.__array__()}\n\n"
            f"number_of_pixels =    {self._number_of_pixels}\n"
            f"pixel_size =          {self._pixel_size}\n\n"
            f"xlim =                {self._xlim}\n"
            f"ylim =                {self._ylim}\n"
            f"pixel_grids:\n"
            f"([{self._pixel_grids[0][0]} ... {self._pixel_grids[0][-1]}],\n"
            f" [{self._pixel_grids[1][0]} ... {self._pixel_grids[1][-1]}])"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPR
        docstr = (
            "Class instance that inherits from `pept.Pixels`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n----------\n"
            f"{self.__str__()}"
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 15.04.2021


'''The `fpi` package implements the Feature Point Identification (FPI)
algorithm for robust voxel-based multiple tracer tracking, using the original
code kindly shared by [1]_.

Summary
-------
A typical workflow for using the `fpi` subpackage would be:

1. Read the LoRs into a `pept.LineData` class instance and set the
   `sample_size` and `overlap` appropriately.
2. Voxellise the `pept.LineData` samples with a `pept.VoxelData` class (this
   can be done on demand, saving memory).
3. Instantiate a `pept.tracking.fpi.FPI` class and transform the voxellised
   LoRs into tracer locations using the `fit` method.

Extended Summary
----------------
[TODO: add more detailed summary of how FPI works].

It was successfully used to track fast-moving radioactive tracers in pipe flows
at the Virginia Commonwealth University. If you use this algorithm in your
work, please cite the original paper [1]_.

Modules Provided
----------------

::

    pept.tracking.fpi
    │
    Functions imported into the subpackage root:
    ├── fpi_ext :   Low-level C++ FPI subroutine.
    │
    Classes imported into the subpackage root:
    └── FPI :       Find tracer locations from samples of voxellised LoRs.

References
----------
.. [1] Wiggins C, Santos R, Ruggles A. A feature point identification method
   for positron emission particle tracking with multiple tracers. Nuclear
   Instruments and Methods in Physics Research Section A: Accelerators,
   Spectrometers, Detectors and Associated Equipment. 2017 Jan 21;843:22-8.
'''


from    .fpi        import  FPI
from    .fpi_ext    import  fpi_ext


__all__ = [
    "FPI",
    "fpi_ext",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : find_minpoints.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 20.10.2020


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np      # import numpy for Python functions
cimport numpy as np     # import numpy for C functions (numpy's C API)


np.import_array()


cdef extern from "find_minpoints_ext.c":
    # C is included here so that it doesn't need to be compiled externally
    pass


cdef extern from "find_minpoints_ext.h":
    double* find_minpoints_ext(
        const double *, const Py_ssize_t, const Py_ssize_t, const Py_ssize_t,
        const double, const double *, const int, Py_ssize_t *, Py_ssize_t *
    ) nogil



cpdef find_minpoints(
    double[:, :] sample_lines,  # LoRs in sample
    Py_ssize_t num_lines,       # Number of LoRs in groups for computing MDP
    double max_distance,        # Max allowed distance between two LoRs
    double[:] cutoffs,          # Spatial cutoff for cutpoints
    bint append_indices = 0     # Append LoR indices used for each cutpoint
):
    '''Compute the minimum distance points (MDPs) from all combinations of
    `num_lines` lines given in an array of lines `sample_lines`.

    ::

        Function signature:
            find_minpoints(
                double[:, :] sample_lines,  # LoRs in sample
                Py_ssize_t num_lines,       # Number of LoRs in combinations
                double max_distance,        # Max distance from MDP to LoRs
                double[:] cutoffs,          # Spatial cutoff for minpoints
                bool append_indices = 0     # Append LoR indices used
            )

    Given a sample of lines, this functions computes the minimum distance
    points (MDPs) for every possible combination of `num_lines` lines. The
    returned numpy array contains all MDPs that satisfy the following:

    1. Are within the `cutoffs`.
    2. Are closer to all the constituent LoRs than `max_distance`.

    Parameters
    ----------
    sample_lines: (M, N) numpy.ndarray
        A 2D array of lines, where each line is defined by two points such that
        every row is formatted as `[t, x1, y1, z1, x2, y2, z2, etc.]`. It
        *must* have at least 2 lines and the combination size `num_lines`
        *must* be smaller or equal to the number of lines. Put differently:
        2 <= num_lines <= len(sample_lines).

    num_lines: int
        The number of lines in each combination of LoRs used to compute the
        MDP. This function considers every combination of `numlines` from the
        input `sample_lines`. It must be smaller or equal to the number of input
        lines `sample_lines`.

    max_distance: float
        The maximum allowed distance between an MDP and its constituent lines.
        If any distance from the MDP to one of its lines is larger than
        `max_distance`, the MDP is thrown away.

    cutoffs: (6,) numpy.ndarray
        An array of spatial cutoff coordinates with *exactly 6 elements* as
        [x_min, x_max, y_min, y_max, z_min, z_max]. If any MDP lies outside
        this region, it is thrown away.

    append_indices: bool
        A boolean specifying whether to include the indices of the lines used
        to compute each MDP. If `False`, the output array will only contain the
        [time, x, y, z] of the MDPs. If `True`, the output array will have
        extra columns [time, x, y, z, line_idx(1), ..., line_idx(n)] where
        n = `num_lines`.

    Returns
    -------
    minpoints: (M, N) numpy.ndarray
        A 2D array of `float`s containing the time and coordinates of the MDPs
        [time, x, y, z]. The time is computed as the average of the constituent
        lines. If `append_indices` is `True`, then `num_lines` indices of the
        constituent lines are appended as extra columns:
        [time, x, y, z, line_idx1, line_idx2, ..].

    Notes
    -----
    There must be at least two lines in `sample_lines` and `num_lines` must be
    greater or equal to the number of lines (i.e. `len(sample_lines)`).
    Put another way: 2 <= num_lines <= len(sample_lines).

    This is a low-level Cython function that does not do any checks on the
    input data - it is meant to be used in other modules / libraries. For a
    normal user, the `pept.tracking.peptml` function `find_minpoints` and
    class `Minpoints` are recommended as higher-level APIs. They do check the
    input data and are easier to use (for example, they automatically compute
    the cutoffs).

    Examples
    --------

    >>> import numpy as np
    >>> from pept.utilities import find_minpoints
    >>>
    >>> lines = np.random.random((500, 7)) * 500
    >>> num_lines = 3
    >>> max_distance = 0.1
    >>> cutoffs = np.array([0, 500, 0, 500, 0, 500], dtype = float)
    >>>
    >>> minpoints = find_minpoints(lines, num_lines, max_distance, cutoffs)

    '''

    # Lines for a single sample => (m, n >= 7) array
    # sample_lines row: [time X1 Y1 Z1 X2 Y2 Z2 etc.]
    cdef Py_ssize_t nrows = sample_lines.shape[0]
    cdef Py_ssize_t ncols = sample_lines.shape[1]

    cdef Py_ssize_t mpts_nrows = 0
    cdef Py_ssize_t mpts_ncols = 0

    cdef double *minpoints
    cdef np.npy_intp[2] size

    with nogil:
        minpoints = find_minpoints_ext(
            &sample_lines[0, 0],
            nrows,
            ncols,
            num_lines,
            max_distance,
            &cutoffs[0],
            append_indices,
            &mpts_nrows,
            &mpts_ncols
        )

    size[0] = mpts_nrows
    size[1] = mpts_ncols

    # Use the `minpoints` pointer as the internal data of a numpy array
    cdef extern from "numpy/arrayobject.h":
        void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

    cdef np.ndarray[double, ndim=2] mpts_arr = np.PyArray_SimpleNewFromData(
        2, size, np.NPY_FLOAT64, minpoints
    )
    PyArray_ENABLEFLAGS(mpts_arr, np.NPY_OWNDATA)

    # Sort rows based on time (column 0)
    mpts_arr = mpts_arr[mpts_arr[:, 0].argsort()]

    return mpts_arr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 21.08.2019


'''Tracer location, identification and tracking algorithms.

Summary
-------
The `pept.tracking` subpackage hosts different tracking algorithms, working
with both the base classes, as well as with generic NumPy arrays. The ones that
use the base classes provide parallelised subroutines, being faster and easier
to use, while the functions using NumPy arrays provide fine-grained control
over the individual samples of lines / points being used.

Subpackages Provided
--------------------

::

    pept.tracking
    │
    Modules imported into the subpackage root:
    ├── birmingham_method :     The Birmingham method for single-tracers.
    ├── peptml :                The PEPT-ML multi-tracer tracking algorithm.
    ├── trajectory_separation : Separate located points into distinct tracks.
    │
    Subpackages:
    ├── birmingham_method
    │   └── BirminghamMethod
    ├── fpi
    │   ├── fpi_ext
    │   └── FPI
    ├── peptml
    │   ├── find_cutpoints
    │   ├── get_cutoffs
    │   ├── Cutpoints
    │   └── HDBSCANClusterer
    └── trajectory_separation
        ├── segregate_trajectories
        ├── connect_trajectories
        └── trajectory_errors

'''


from    .   import  birmingham_method
from    .   import  peptml
from    .   import  fpi
from    .   import  trajectory_separation


__all__ = [
    "birmingham_method",
    "peptml",
    "fpi",
    "trajectory_separation"
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# MIT License
#
# Copyright (c) 2016 Thomas Baruchel
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Allow to use tail-call optimized functions in Python code (for
tail-recursion or continuation-passing style).
"""

# This script was taken by Andrei Leonard Nicusan in 2020 from Thomas
# Baruchel's GitHub repository at:
# https://github.com/baruchel/tco
#
# At the README.md's suggestion, the whole library (consisting of a single
# __init__.py file) was copied into this file for ease of use, avoiding the
# addition of another dependency to the pept package.
#
# I express my thanks to Thomas Baruchel for this beautifully minimalistic
# implementation of ad hoc tail call optimisation for Python.

__version__ = '1.2.1'

class _TailCall(Exception):
    def __init__(self, f, args, uid):
        self.func, self.args, self.uid, self.follow = f.func, args, uid, id(f)

def _tailCallback(f, uid):
    """
    This is the "callable" version of the continuation, which sould only
    be accessible from the inside of the function to be continued. An
    attribute called "C" can be used in order to get back the public
    version of the continuation (for passing the continuation to another
    function).
    """
    def t(*args):
        raise _TailCall(f, args, uid)
    t.C = f
    return t

class _TailCallWrapper():
    """
    Wrapper for tail-called optimized functions embedding their
    continuations. Such functions are ready to be evaluated with
    their arguments.
    This is a private class and should never be accessed directly.
    Functions should be created by using the C() class first.
    """
    def __init__(self, func, k):
        self.func = func( _tailCallback(self, id(self)),
                          *map( lambda c: _tailCallback(c, id(self)), k) )
    def __call__(self, *args):
        f, expect = self.func, id(self)
        while True:
            try:
                return f(*args)
            except _TailCall as e:
                if e.uid == expect:
                    f, args, expect = e.func, e.args, e.follow
                else:
                    raise e

class C():
    """
    Main wrapper for tail-call optimized functions.
    """
    def __init__(self, func):
        self.func = func
    def __call__(self, *k):
        return _TailCallWrapper(self.func, k)

def with_continuations(**c):
    """
    A decorator for defining tail-call optimized functions.
    Example
    -------
        @with_continuations()
        def factorial(n, k, self=None):
            return self(n-1, k*n) if n > 1 else k

        @with_continuations()
        def identity(x, self=None):
            return x

        @with_continuations(out=identity)
        def factorial2(n, k, self=None, out=None):
            return self(n-1, k*n) if n > 1 else out(k)
        print(factorial(7,1))
        print(factorial2(7,1))
    """
    if len(c): keys, k = zip(*c.items())
    else: keys, k = tuple([]), tuple([])
    def d(f):
        return C(
            lambda kself, *conts:
                lambda *args:
                    f(*args, self=kself, **dict(zip(keys, conts)))) (*k)
    return d



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : cutpoints.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 13.04.2020


import  time
import  os
import  warnings
import  textwrap

import  numpy               as      np

from    tqdm                import  tqdm

from    concurrent.futures  import  ThreadPoolExecutor

import  pept


def find_cutpoints(
    sample_lines,
    max_distance,
    cutoffs = None,
    append_indices = False
):
    '''Find the cutpoints from a sample / array of LoRs.

    A cutpoint is the point in 3D space that minimises the distance between any
    two lines. For any two non-parallel 3D lines, this point corresponds to the
    midpoint of the unique segment that is perpendicular to both lines.

    This function considers every pair of lines in `sample_lines` and returns
    all the cutpoints that satisfy the following conditions:

    1. The distance between the two lines is smaller than `max_distance`.
    2. The cutpoint is within the `cutoffs`.

    Parameters
    ----------
    sample_lines : (N, M >= 7) numpy.ndarray
        A sample of LoRs, where each row is `[time, x1, y1, z1, x2, y2, z2]`,
        such that every line is defined by the points `[x1, y1, z1]` and
        `[x2, y2, z2]`.
    max_distance : float
        The maximum distance between any two lines for their cutpoint to be
        considered. A good starting value would be 0.1 mm for small tracers
        and/or clean data, or 0.2 mm for larger tracers and/or noisy data.
    cutoffs : list, optional
        The cutoffs for each dimension, formatted as `[x_min, x_max,
        y_min, y_max, z_min, z_max]`. If it is `None`, they are computed
        automatically by calling `get_cutoffs`. The default is `None`.
    append_indices : bool, optional
        If set to `True`, the indices of the individual LoRs that were used
        to compute each cutpoint are also appended to the returned array.
        Default is `False`.

    Returns
    -------
    cutpoints : (M, 4) or (M, 6) numpy.ndarray
        A numpy array of the calculated cutpoints. If `append_indices` is
        `False`, then the columns are [time, x, y, z]. If `append_indices` is
        `True`, then the columns are [time, x, y, z, i, j], where `i` and `j`
        are the LoR indices from `sample_lines` that were used to compute the
        weighted cutpoints. The time is the average between the timestamps of
        the two LoRs that were used to compute the cutpoint. The first column
        (for time) is sorted.

    Raises
    ------
    ValueError
        If `sample_lines` is not a numpy array with shape (N, M >= 7).
    ValueError
        If `cutoffs` is not a one-dimensional array with values
        `[min_x, max_x, min_y, max_y, min_z, max_z]`

    See Also
    --------
    pept.tracking.peptml.Cutpoints : Compute cutpoints from `pept.LineData`.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    '''

    sample_lines = np.asarray(sample_lines, order = 'C', dtype = float)
    max_distance = float(max_distance)

    # Check sample has shape (N, M >= 7)
    if sample_lines.ndim != 2 or sample_lines.shape[1] < 7:
        raise ValueError((
            "\n[ERROR]: `sample_lines` should have dimensions (M, N), "
            f" where N >= 7. Received {sample_lines.shape}.\n"
        ))

    if cutoffs is None:
        cutoffs = get_cutoffs(sample_lines)
    else:
        cutoffs = np.asarray(cutoffs, order = 'C', dtype = float)
        if cutoffs.ndim != 1 or len(cutoffs) != 6:
            raise ValueError((
                "\n[ERROR]: cutoffs should be a one-dimensional array with "
                "values [min_x, max_x, min_y, max_y, min_z, max_z]. Received "
                f"{cutoffs}.\n"
            ))

    sample_cutpoints = pept.utilities.find_cutpoints(
        sample_lines,
        max_distance,
        cutoffs,
        append_indices = append_indices
    )

    return sample_cutpoints


def get_cutoffs(sample):
    '''Compute the cutoffs from a sample of LoR data.

    It computes the cutoffs from the minimum and maximum values of the LoRs in
    `sample` in each dimension (e.g. the x-dimension is defined by data in
    columns 1 and 4).

    Parameters
    ----------
    sample : (N, M >= 7) numpy.ndarray
        A sample of LoRs, where each row is `[time, x1, y1, z1, x2, y2, z2]`,
        such that every line is defined by the points `[x1, y1, z1]` and
        `[x2, y2, z2]`.

    Returns
    -------
    cutoffs : (6,) numpy.ndarray
        The computed cutoffs for each dimension, formatted as
        `[x_min, x_max, y_min, y_max, z_min, z_max]`.

    Raises
    ------
    ValueError
        If `sample` is not a numpy array with shape (N, M >= 7).

    See Also
    --------
    pept.tracking.peptml.Cutpoints : Compute cutpoints from `pept.LineData`.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    '''

    # Check sample has shape (N, M >= 7)
    if sample.ndim != 2 or sample.shape[1] < 7:
        raise ValueError((
            "\n[ERROR]: `sample_lines` should have dimensions (M, N), "
            f" where N >= 7. Received {sample.shape}.\n"
        ))

    # Compute cutoffs for cutpoints as the (min, max) values of the lines
    # Minimum value of the two points that define a line
    min_x = min(sample[:, 1].min(),
                sample[:, 4].min())
    # Maximum value of the two points that define a line
    max_x = max(sample[:, 1].max(),
                sample[:, 4].max())

    # Minimum value of the two points that define a line
    min_y = min(sample[:, 2].min(),
                sample[:, 5].min())
    # Maximum value of the two points that define a line
    max_y = max(sample[:, 2].max(),
                sample[:, 5].max())

    # Minimum value of the two points that define a line
    min_z = min(sample[:, 3].min(),
                sample[:, 6].min())
    # Maximum value of the two points that define a line
    max_z = max(sample[:, 3].max(),
                sample[:, 6].max())

    cutoffs = np.array([min_x, max_x, min_y, max_y, min_z, max_z],
                       dtype = float)
    return cutoffs




class Cutpoints(pept.PointData):
    '''Transform LoRs (a `pept.LineData` instance) into *cutpoints* (a
    `pept.PointData` instance) for clustering, in parallel.

    Under typical usage, the `Cutpoints` class is initialised with a
    `pept.LineData` instance, automatically calculating the cutpoints from the
    samples of lines. The `Cutpoints` class inherits from `pept.PointData`,
    such that once the cutpoints have been computed, all the methods from the
    parent class `pept.PointData` can be used on them (such as visualisation
    functionality).

    For more control over the operations, `pept.tracking.peptml.find_cutpoints`
    can be used - it receives a generic numpy array of LoRs (one 'sample') and
    returns a numpy array of cutpoints.

    Attributes
    ----------
    line_data : instance of pept.LineData
        The LoRs for which the cutpoints will be computed. It must be an
        instance of `pept.LineData`.
    max_distance : float
        The maximum distance between any two lines for their cutpoint to be
        considered. A good starting value would be 0.1 mm for small tracers
        and/or clean data, or 0.2 mm for larger tracers and/or noisy data.
    cutoffs : list-like of length 6
        A list (or equivalent) of the cutoff distances for every axis,
        formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
        cutpoints which fall within these cutoff distances are considered. The
        default is None, in which case they are automatically computed using
        `pept.tracking.peptml.get_cutoffs`.
    sample_size, overlap, number_of_lines, etc. : inherited from pept.PointData
        Additional attributes and methods are inherited from the base class
        `PointData`. Check its documentation for more information.

    Methods
    -------
    find_cutpoints(line_data, max_distance, cutoffs = None,\
                   append_indices = False, max_workers = None, verbose = True)
        Compute the cutpoints from the samples in a `LineData` instance.
    sample, to_csv, plot, etc. : inherited from pept.PointData
        Additional attributes and methods are inherited from the base class
        `PointData`. Check its documentation for more information.

    Notes
    -----
    Once instantiated with a `LineData`, the class computes the cutpoints and
    *automatically sets the sample_size* to the average number of cutpoints
    found per sample of LoRs.

    Examples
    --------
    Compute the cutpoints for a `LineData` instance between lines that are less
    than 0.1 apart:

    >>> line_data = pept.LineData(example_data)
    >>> cutpts = peptml.Cutpoints(line_data, 0.1)

    Compute the cutpoints for a single sample:

    >>> sample = line_data[0]
    >>> cutpts_sample = peptml.find_cutpoints(sample, 0.1)

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.tracking.peptml.HDBSCANClusterer : Efficient, parallel HDBSCAN-based
                                            clustering of cutpoints.
    pept.scanners.ParallelScreens : Read in and initialise a `pept.LineData`
                                    instance from parallel screens PET/PEPT
                                    detectors.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    '''

    def __init__(
        self,
        line_data,
        max_distance,
        cutoffs = None,
        append_indices = False,
        max_workers = None,
        verbose = True
    ):
        '''Cutpoints class constructor.

        Parameters
        ----------
        line_data : instance of pept.LineData
            The LoRs for which the cutpoints will be computed. It must be an
            instance of `pept.LineData`.
        max_distance : float
            The maximum distance between any two lines for their cutpoint to be
            considered. A good starting value would be 0.1 mm for small tracers
            and/or clean data, or 0.2 mm for larger tracers and/or noisy data.
        cutoffs : list-like of length 6, optional
            A list (or equivalent) of the cutoff distances for every axis,
            formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
            cutpoints which fall within these cutoff distances are considered.
            The default is None, in which case they are automatically computed
            using `pept.tracking.peptml.get_cutoffs`.
        append_indices : bool, default False
            If set to `True`, the indices of the individual LoRs that were used
            to compute each cutpoint are also appended to the returned array.
        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            computing the cutpoints from the samples of LoRs in `line_data`.
        verbose : bool, default True
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        Raises
        ------
        TypeError
            If `line_data` is not an instance of `pept.LineData`.
        ValueError
            If `cutoffs` is not a one-dimensional array with values formatted
            as `[min_x, max_x, min_y, max_y, min_z, max_z]`.
        '''

        # Find the cutpoints when instantiated. The method
        # also initialises the instance as a `PointData` subclass.
        self.find_cutpoints(
            line_data,
            max_distance,
            cutoffs = cutoffs,
            append_indices = append_indices,
            max_workers = max_workers,
            verbose = verbose
        )


    @property
    def line_data(self):
        return self._line_data


    @property
    def max_distance(self):
        return self._max_distance


    @property
    def cutoffs(self):
        return self._cutoffs


    def find_cutpoints(
        self,
        line_data,
        max_distance,
        cutoffs = None,
        append_indices = False,
        max_workers = None,
        verbose = True
    ):
        '''Compute the cutpoints from the samples in a `LineData` instance.

        Parameters
        ----------
        line_data : instance of pept.LineData
            The LoRs for which the cutpoints will be computed. It must be an
            instance of `pept.LineData`.
        max_distance : float
            The maximum distance between any two LoRs for their cutpoint to be
            considered.
        cutoffs : list-like of length 6, optional
            A list (or equivalent) of the cutoff distances for every axis,
            formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
            cutpoints which fall within these cutoff distances are considered.
            The default is None, in which case they are automatically computed
            using `pept.tracking.peptml.get_cutoffs`.
        append_indices : bool, default False
            If set to `True`, the indices of the individual LoRs that were used
            to compute each cutpoint are also appended to the returned array.
        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            computing the cutpoints from the samples of LoRs in `line_data`.
        verbose : bool, default True
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        Returns
        -------
        self : the PointData instance of cutpoints
            The computed cutpoints are stored in the `Cutpoints` class, as a
            subclass of `pept.PointData`.

        Raises
        ------
        TypeError
            If `line_data` is not an instance of `pept.LineData`.
        ValueError
            If `cutoffs` is not a one-dimensional array with values formatted
            as `[min_x, max_x, min_y, max_y, min_z, max_z]`.

        Notes
        -----
        This method is automatically called when instantiating this class.
        '''

        if verbose:
            start = time.time()

        # Check line_data is an instance (or a subclass!) of pept.LineData
        if not isinstance(line_data, pept.LineData):
            raise TypeError((
                "\n[ERROR]: line_data should be an instance (or subclass) of "
                "`pept.LineData`.\n"
            ))

        # Users might forget to set the sample_size, leaving it to the default
        # value of 0; in that case, all lines are returned as a single sample -
        # that might not be the intended behaviour.
        if line_data.sample_size == 0:
            warnings.warn(
                textwrap.fill((
                    "\n[WARNING]: The `line_data.sample_size` was left to the "
                    "default value of 0, in which case all lines are returned "
                    "as a single sample. For a very large number of lines, "
                    "this might result in a long function execution time.\n"
                ), replace_whitespace = False),
                RuntimeWarning
            )

        self._line_data = line_data
        self._max_distance = float(max_distance)

        # If cutoffs were not supplied, compute them
        if cutoffs is None:
            cutoffs = get_cutoffs(line_data.lines)
        # Otherwise make sure they are a C-contiguous numpy array
        else:
            cutoffs = np.asarray(cutoffs, order = 'C', dtype = float)
            if cutoffs.ndim != 1 or len(cutoffs) != 6:
                raise ValueError((
                    "\n[ERROR]: cutoffs should be a one-dimensional array "
                    "with values [min_x, max_x, min_y, max_y, min_z, max_z]. "
                    f"Received {cutoffs}.\n"
                ))

        self._cutoffs = cutoffs

        # Using ThreadPoolExecutor, asynchronously collect the cutpoints from
        # every sample in a list of arrays. This is more efficient than using
        # ProcessPoolExecutor because find_cutpoints is a Cython function that
        # releases the GIL for most of its computation.
        # If verbose, show progress bar using tqdm.
        if max_workers is None:
            max_workers = os.cpu_count()

        with ThreadPoolExecutor(max_workers = max_workers) as executor:
            futures = []
            for sample in line_data:
                futures.append(
                    executor.submit(
                        pept.utilities.find_cutpoints,
                        sample,
                        max_distance,
                        cutoffs,
                        append_indices = append_indices
                    )
                )

            if verbose:
                futures = tqdm(futures)

            cutpoints = [f.result() for f in futures]

        # cutpoints shape: (n, m, 4), where n is the number of samples, and
        # m is the number of cutpoints in the sample
        number_of_samples = len(cutpoints)
        cutpoints = np.vstack(cutpoints)
        number_of_cutpoints = len(cutpoints)

        # Average number of cutpoints per sample
        cutpoints_per_sample = int(number_of_cutpoints / number_of_samples)

        pept.PointData.__init__(
            self,
            cutpoints,
            sample_size = cutpoints_per_sample,
            overlap = 0,
            verbose = False
        )

        if verbose:
            end = time.time()
            print(f"\nFinding the cutpoints took {end - start} seconds.\n")

        return self


    def __repr__(self):
        # Called when writing the class on a REPL. Add another line to the
        # standard description given in the parent class, pept.PointData.
        docstr = pept.PointData.__repr__(self) + (
            "\n\nNotes\n-----\n"
            "Once instantiated with a `LineData`, the class computes the \n"
            "cutpoints and *automatically sets the sample_size* to the \n"
            "average number of cutpoints found per sample of LoRs."
        )

        return docstr



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : find_cutpoints.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 03.02.2019


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np


cpdef find_cutpoints(
    double[:, :] sample_lines,  # LoRs in sample
    double max_distance,        # Max allowed distance between two LoRs
    double[:] cutoffs,          # Spatial cutoff for cutpoints
    bint append_indices = 0     # Append LoR indices used for each cutpoint
):
    '''Compute the cutpoints from a given array of lines.

    ::

        Function signature:
            find_cutpoints(
                double[:, :] sample_lines,  # LoRs in sample
                double max_distance,        # Max distance between two LoRs
                double[:] cutoffs,          # Spatial cutoff for cutpoints
                bint append_indices = False # Append LoR indices used
            )

    This is a low-level Cython function that does not do any checks on the
    input data - it is meant to be used in other modules / libraries. For a
    normal user, the `pept.tracking.peptml` function `find_cutpoints` and
    class `Cutpoints` are recommended as higher-level APIs. They do check the
    input data and are easier to use (for example, they automatically compute
    the cutoffs).

    A cutpoint is the point in 3D space that minimises the distance between any
    two lines. For any two non-parallel 3D lines, this point corresponds to the
    midpoint of the unique segment that is perpendicular to both lines.

    This function considers every pair of lines in `sample_lines` and returns
    all the cutpoints that satisfy the following conditions:

    1. The distance between the two lines is smaller than `max_distance`.
    2. The cutpoints are within the `cutoffs`.

    Parameters
    ----------
    sample_lines : (N, M >= 7) numpy.ndarray
        The sample of lines, where each row is [time, x1, y1, z1, x2, y2, z2],
        containing two points [x1, y1, z1] and [x2, y2, z2] defining an LoR.
    max_distance : float
        The maximum distance between two LoRs for their cutpoint to be
        considered.
    cutoffs : (6,) numpy.ndarray
        Only consider the cutpoints that fall within the cutoffs. `cutoffs` has
        the format [min_x, max_x, min_y, max_y, min_z, max_z].
    append_indices : bool, optional
        If set to `True`, the indices of the individual LoRs that were used
        to compute each cutpoint is also appended to the returned array.
        Default is `False`.

    Returns
    -------
    cutpoints : (M, 4) or (M, 6) numpy.ndarray
        A numpy array of the calculated weighted cutpoints. If `append_indices`
        is `False`, then the columns are [time, x, y, z]. If `append_indices`
        is `True`, then the columns are [time, x, y, z, i, j], where `i` and
        `j` are the LoR indices from `sample_lines` that were used to compute
        the cutpoints. The time is the average between the timestamps of the
        two LoRs that were used to compute the cutpoint. The first column (for
        time) is sorted.

    Examples
    --------

    >>> import numpy as np
    >>> from pept.utilities import find_cutpoints
    >>>
    >>> lines = np.random.random((500, 7)) * 500
    >>> max_distance = 0.1
    >>> cutoffs = np.array([0, 500, 0, 500, 0, 500], dtype = float)
    >>>
    >>> cutpoints = find_cutpoints(lines, max_distance, cutoffs)

    '''

    # Lines for a single sample => n x 7 array
    # sample_lines row: [time X1 Y1 Z1 X2 Y2 Z2]
    cdef Py_ssize_t n = sample_lines.shape[0]

    # Pre-allocate enough memory
    # weighted_cutpoints cols: [time, x, y, z, [LoR1_index, LoR2_index]]
    if append_indices:
        cutpoints = np.zeros((n * (n - 1) // 2, 6), order = 'C')
    else:
        cutpoints = np.zeros((n * (n - 1) // 2, 4), order = 'C')


    cdef double[:, :] cutpoints_mv = cutpoints  # memoryview of cutpoints
    cdef Py_ssize_t i, j, k                     # iterators
    cdef Py_ssize_t ic                          # cutpoint index
    cdef double[3] P, U, Q, R, QP               # position, direction vectors
    cdef double a, b, c, d, e                   # collected terms
    cdef double denom, s0, t0                   # parameters for lines
    cdef double[3] A0, C0, AC0                  # perpendicular points
    cdef double mx, my, mz                      # cutpoint coordinates

    with nogil:
        ic = 0
        for i in range(n - 1):
            for j in range(i + 1, n):

                # Write each pair of lines in terms of a position vector and a
                # direction vector:
                # L1 : A(s) = P + s U
                # L2 : C(t) = Q + t R
                for k in range(3):
                    P[k] = sample_lines[i, 1 + k]
                    U[k] = sample_lines[i, 4 + k] - P[k]
                    Q[k] = sample_lines[j, 1 + k]
                    R[k] = sample_lines[j, 4 + k] - Q[k]
                    QP[k] = Q[k] - P[k]

                a = U[0] * U[0] + U[1] * U[1] + U[2] * U[2]
                b = U[0] * R[0] + U[1] * R[1] + U[2] * R[2]
                c = R[0] * R[0] + R[1] * R[1] + R[2] * R[2]
                d = U[0] * QP[0] + U[1] * QP[1] + U[2] * QP[2]
                e = QP[0] * R[0] + QP[1] * R[1] + QP[2] * R[2]

                denom = b * b - a * c
                if denom:
                    s0 = (b * e - c * d) / denom
                    t0 = (a * e - b * d) / denom

                    for k in range(3):
                        A0[k] = P[k] + s0 * U[k]
                        C0[k] = Q[k] + t0 * R[k]
                        AC0[k] = A0[k] - C0[k]

                    # Check the distance is smaller than max_distance
                    if (AC0[0] * AC0[0] + AC0[1] * AC0[1] +
                        AC0[2] * AC0[2] < max_distance * max_distance):
                        mx = (A0[0] + C0[0]) / 2
                        my = (A0[1] + C0[1]) / 2
                        mz = (A0[2] + C0[2]) / 2

                        # Check the cutpoint falls within the cutoffs
                        if (mx > cutoffs[0] and mx < cutoffs[1] and
                            my > cutoffs[2] and my < cutoffs[3] and
                            mz > cutoffs[4] and mz < cutoffs[5]):
                            # Average the times of the two lines
                            cutpoints_mv[ic, 0] = (sample_lines[i, 0] + \
                                                   sample_lines[j, 0]) / 2
                            cutpoints_mv[ic, 1] = mx
                            cutpoints_mv[ic, 2] = my
                            cutpoints_mv[ic, 3] = mz

                            if append_indices:
                                cutpoints_mv[ic, 4] = <double>i
                                cutpoints_mv[ic, 5] = <double>j

                            ic = ic + 1


    # Truncate the cutpoints which were not written
    cutpoints_truncated = np.delete(cutpoints, slice(ic, None, None), 0)
    del(cutpoints)

    # Sort rows based on time (column 0)
    cutpoints_truncated = cutpoints_truncated[cutpoints_truncated[:, 0].argsort()]

    return cutpoints_truncated



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.08.2019


from    .parallel_screens   import  ParallelScreens
from    .adac_forte         import  ADACForte
from    .extensions         import  convert_adac_forte


__all__ = [
    "ParallelScreens",
    "ADACForte",
    "convert_adac_forte",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.06.2020


'''The PEPT cookbook, containing example analysis scripts (or recipes) that are
easy to hack into.

Summary
-------
Following the cookbook metaphor, this package contains a collection of scripts
combining different parts of the `pept` library (i.e. ingredients) to
accomplish a certain task -> recipes! Much like when cooking, it is more than
recommended to take the recipes (i.e. code) and make them yours; they are
well-commented and user-friendly.

Extended Summary
----------------
The recipes are implemented *as classes*, such that once they've been run (i.e.
instantiated), all the relevant objects (such as clusterers, graphers, etc.)
can be accessed as class attributes.

Modules Provided
----------------

::

    pept.cookbook
    │
    Classes imported into the subpackage root:
    ├── PEPTMLUser :            Use PEPT-ML to turn LoRs into trajectories.
    └── PEPTMLFindParameters :  Visual aid for PEPT-ML clustering parameters.

Notes
-----
At the moment, the subpackages in `pept` are biased towards PEPT-ML, as there
aren't many algorithms integrated into package *yet*. New algorithms and/or
recommendations for the package are more than welcome! `pept` aims to be a
community effort, be it academic, industrial, medical, or just from PEPT
enthusiasts - so it is open to help with documentation, algorithms, utilities
or analysis scripts, tutorials, and pull requests in general!
'''


from    .peptml     import  PEPTMLUser
from    .peptml     import  PEPTMLFindParameters


__all__ = [
    "PEPTMLUser",
    "PEPTMLFindParameters"
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"
